# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Module containing the main plotting code for AlphaGenome model outputs.

Three main elements are:

  * `plot` function: The primary function for visualizing model outputs.
  * Component classes: Implement the visualization components (e.g., tracks,
    contact maps).
  * Annotation classes: Implement the visualization annotations (e.g.,
    intervals, variants).
"""

import abc
from collections.abc import Mapping, Sequence

from alphagenome.data import genome
from alphagenome.data import junction_data
from alphagenome.data import track_data
from alphagenome.data import transcript as transcript_utils
from alphagenome.visualization import plot as plot_lib
from alphagenome.visualization import plot_transcripts
from jaxtyping import Float32  # pylint: disable=g-importing-member
import matplotlib
from matplotlib import colors as plt_colors
import matplotlib.pyplot as plt
import numpy as np


# String, RGB or RGBA color.
_ColorType = (
    str | tuple[float, float, float] | tuple[float, float, float, float]
)


def plot(
    components: Sequence['AbstractComponent'],
    interval: genome.Interval,
    fig_width: int = 20,
    fig_height_scale: float = 1.0,
    title: str | None = None,
    despine: bool = True,
    despine_keep_bottom: bool = False,
    annotations: Sequence['AbstractAnnotation'] | None = None,
    annotation_offset_range: tuple[float, float] = (0.1, 0.6),
    hspace: float = 0.3,
    xlabel: str | None = None,
) -> matplotlib.figure.Figure:
  """Plots AlphaGenome model outputs as individual panels of 'components'.

  This function generates a visualization of AlphaGenome model outputs
  using a combination of components (e.g., tracks, contact maps) and
  annotations (e.g., intervals, variants).

  Args:
    components: A sequence of components to visualize.
    interval: The genomic interval to focus on (similar to setting xlim in
      matplotlib).
    fig_width: The total figure width.
    fig_height_scale: Height of the individual track unit. Total plot height is
      determined as a sum of the individual component heights.
    title: An optional title for the overall plot.
    despine: Whether to remove top, right, and bottom spines from the axes.
    despine_keep_bottom: Whether to remove top and right spines, but keep the
      bottom spine. Does not apply to transcript components, which are always
      despined.
    annotations: Sequence of annotations to visualise across all components.
    annotation_offset_range: Relative positions in y-axis to place labels for
      annotations. Each set of labels in the 'annotations' list are spaced
      evenly within this range.
    hspace: Vertical whitespace between subplots to avoid tick label overlap
      (relative fraction).
    xlabel: If a non-empty string is provided, it is used as the x-axis label.
      If an empty string is provided, the x-axis label is removed. If None, the
      default x-axis label is used.

  Returns:
    A matplotlib figure.
  """

  # If we are adding text labels for any of the annotation components,
  # we need to add an extra empty component at the top.
  add_label_axis = False
  if annotations is not None:
    add_label_axis = any(annot.has_labels for annot in annotations)
    components = (
        [EmptyComponent()] + list(components)
        if add_label_axis
        else list(components)
    )

  num_axes = sum(component.num_axes for component in components)
  fig_height = sum(
      component.total_height * fig_height_scale for component in components
  )

  offset = 0
  gridspec_kw = {'height_ratios': []}
  for component in components:
    for i in range(component.num_axes):
      gridspec_kw['height_ratios'].append(
          component.get_ax_height(i) * fig_height_scale / fig_height
      )
      offset += 1

  fig, axes = plt.subplots(
      nrows=num_axes,
      ncols=1,
      figsize=(fig_width, fig_height),
      gridspec_kw=gridspec_kw,
      sharex=True,
  )
  # Handle the case of a single axis (e.g. single REF / ALT plot).
  if not isinstance(axes, np.ndarray):
    axes = [axes]

  offset = 0
  for component in components:
    for i in range(component.num_axes):
      ax = axes[offset]
      component.plot_ax(ax, axis_index=i, interval=interval)

      if offset == 0 and title is not None:
        ax.set_title(title)

      if despine:
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        if (
            i != num_axes - 1
            and despine_keep_bottom
            and (not isinstance(component, TranscriptAnnotation))
        ):
          ax.spines['bottom'].set_visible(True)
        else:
          ax.spines['bottom'].set_visible(False)

      # Remove tick marks which become visible due to a subplots_adjust() call.
      if offset != num_axes - 1:
        ax.tick_params(axis='x', which='both', bottom=False, labelbottom=False)

      ax.set_xlim(interval.start, interval.end)

      offset += 1

  # Add annotations across all plotted components.
  if annotations is not None:
    # Add small vertical offset for each set of labels to avoid overlap.
    num_labelled_annotations = sum(
        annotation.has_labels for annotation in annotations
    )
    height_offsets = np.linspace(
        annotation_offset_range[0],
        annotation_offset_range[1],
        num_labelled_annotations,
    )
    # Identify which axes are transcripts and which we want to annotate.
    transcript_axes_idx = [
        i
        for i, comp in enumerate(components)
        if isinstance(comp, TranscriptAnnotation)
    ]
    axes_to_annotate_idx = (
        range(1, len(axes)) if add_label_axis else range(len(axes))
    )
    for i, annotation in enumerate(annotations):
      for j in axes_to_annotate_idx:
        if not annotation.is_variant and j in transcript_axes_idx:
          # Do not add interval annotations to axes that involves transcripts.
          continue
        annotation.plot_ax(axes[j], interval, hspace)
      if annotation.has_labels:
        # Add labels to the empty axis at the top. All labels are added to
        # the top axis, in the order they are supplied in annotations list.
        annotation.plot_labels(axes[0], interval, height_offsets[i])

  # Enable default tick locator for the final subplot.
  axes[-1].xaxis.set_major_locator(matplotlib.ticker.AutoLocator())

  if xlabel is not None:
    axes[-1].set_xlabel(xlabel)
  else:
    axes[-1].set_xlabel(f'Chromosome position; interval={interval}')

  # Slight whitespace between subplots to avoid tick label overlap.
  fig.subplots_adjust(hspace=hspace)

  return fig


class AbstractComponent(abc.ABC):
  """Abstract base class for plot components."""

  @abc.abstractmethod
  def get_ax_height(self, axis_index: int) -> float:
    """Returns the plot height for the individual axis.

    Args:
      axis_index: The index of the axis.

    Returns:
      The height of the axis.
    """

  @property
  def total_height(self) -> float:
    """Returns the total figure height."""
    return sum(self.get_ax_height(i) for i in range(self.num_axes))

  @property
  @abc.abstractmethod
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""

  @abc.abstractmethod
  def plot_ax(
      self,
      ax: matplotlib.axes.Axes,
      axis_index: int,
      interval: genome.Interval,
  ):
    """Plots the component on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """


class Tracks(AbstractComponent):
  """Component for visualizing tracks."""

  def __init__(
      self,
      tdata: track_data.TrackData,
      cmap: str = 'viridis',
      truncate_cmap: bool = True,
      track_height: float = 1.0,
      filled: bool = False,
      ylabel_template: str = '{name}:{strand}',
      ylabel_horizontal: bool = True,
      shared_y_scale: bool = False,
      global_ylims: tuple[float, float] | None = None,
      max_num_tracks: int = 50,
      track_colors: Sequence[_ColorType] | str | None = None,
      **kwargs,
  ):
    """Initializes the `Tracks` component.

    Args:
      tdata: The `TrackData` object to visualize.
      cmap: The colormap to use for the tracks.
      truncate_cmap: Whether to slightly truncate the colormap to avoid extreme
        values.
      track_height: The height of each track.
      filled: Whether to fill the area under the tracks.
      ylabel_template: A template for the y-axis labels.
      ylabel_horizontal: Whether to make the y-axis labels horizontal.
      shared_y_scale: Whether to use the same y-axis scale for all tracks.
      global_ylims: Optional global y-axis limits (min and max).
      max_num_tracks: The maximum number of tracks to plot.
      track_colors: An optional sequence of colors to use for the tracks. If a
        string is passed, it is used as a single color for all tracks.
      **kwargs: Additional keyword arguments to pass to the plotting function.

    Raises:
      ValueError: If the number of tracks exceeds `max_num_tracks` or if the
        track data has more than one positional axis or if the interval is not
        set within the track data.
    """
    if tdata.num_tracks > max_num_tracks:
      raise ValueError(
          f'Too many tracks to plot: {tdata.num_tracks} > {max_num_tracks}.'
      )
    self._tdata = tdata
    self._num_tracks = tdata.values.shape[-1]
    self._track_height = track_height
    self._filled = filled
    self._ylabel_horizontal = ylabel_horizontal
    self._ylabel_template = ylabel_template
    self._shared_y_scale = shared_y_scale
    self._global_ylims = global_ylims
    self._kwargs = kwargs

    if len(self._tdata.positional_axes) != 1:
      raise ValueError(
          'Only track_data with 1 positional axis is supported in Tracks.'
      )
    if self._tdata.interval is None:
      raise ValueError('.interval needs to be set in track_data.')

    # Set up color per set of interleaved tracks.
    if getattr(self._tdata, 'uns') is not None:
      self._num_tdata = self._tdata.uns['num_interleaved_trackdatas']
    else:
      self._num_tdata = 1

    cmap = plt.get_cmap(cmap)
    num_track_sets = self._num_tracks // self._num_tdata
    if truncate_cmap:
      # We do *1.2 to make the color change more gradual. This means that the
      # the upper range of the cmap is never displayed, which often tends to
      # achieve nicer results aesthetically.
      self._colors = cmap(np.linspace(0, 1, round(num_track_sets * 1.2)))
    else:
      self._colors = cmap(np.linspace(0, 1, num_track_sets))

    if track_colors is not None:
      if isinstance(track_colors, str):
        self._colors = [track_colors] * num_track_sets
      elif len(track_colors) == 1:
        self._colors = track_colors * num_track_sets
      elif len(track_colors) != num_track_sets:
        raise ValueError(
            f'track_colors argument (length: {len(track_colors)}) must be'
            ' either a single color, or the same number of track sets provided'
            f' in the tdata ({num_track_sets}).'
        )
      else:
        self._colors = track_colors

  def _get_ylimits(self, tdata: track_data.TrackData):
    """Computes y-axis limits for track sets."""
    return [
        (
            tdata.values[:, i : i + self._num_tdata].min(),
            tdata.values[:, i : i + self._num_tdata].max(),
        )
        for i in range(0, self._num_tracks, self._num_tdata)
    ]

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._track_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    return self._tdata.num_tracks

  def plot_ax(
      self,
      ax: matplotlib.axes.Axes,
      axis_index: int,
      interval: genome.Interval,
  ):
    """Plots the tracks on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """
    tdata = self._tdata
    assert tdata.interval is not None

    # If an interval is passed, zoom in on that specific sub-interval.
    if interval is not None:
      tdata = tdata.slice_by_interval(interval, match_resolution=True)
      del interval
    x = (
        np.arange(tdata.values.shape[0]) * tdata.resolution
        + tdata.interval.start
        + tdata.resolution / 2
    )
    arr = tdata.values[:, axis_index]

    # Same shared y-axis limits across all tracks.
    if self._shared_y_scale:
      if self._global_ylims is not None:
        ax.set_ylim(self._global_ylims)
      else:
        ax.set_ylim(tdata.values.min(), tdata.values.max())
    else:
      ax.set_ylim(arr.min(), arr.max())

    # Set the colour of this track.
    track_color = self._colors[axis_index // self._num_tdata]

    # Draw the line-plot, or filled line-plot if filled=True.
    if self._filled:
      ax.fill_between(x, np.ravel(arr), color=track_color, **self._kwargs)
    else:
      ax.plot(x, arr, c=track_color, **self._kwargs)

    if self._ylabel_template:
      _set_ylabel(ax, self._get_ylabel(axis_index), self._ylabel_horizontal)

  def _get_ylabel(self, axis_index: int) -> str:
    """Returns the y-axis label for the given axis index."""
    row = self._tdata.metadata.iloc[axis_index]
    return self._ylabel_template.format(**row.to_dict())


class OverlaidTracks(AbstractComponent):
  """Component for visualizing overlaid track pairs, such as REF/ALT tracks."""

  def __init__(
      self,
      tdata: Mapping[str, track_data.TrackData],
      colors: Mapping[str, str] | None = None,
      cmap: str | None = 'viridis',
      track_height: float = 1.0,
      ylabel_template: str = '{name}:{strand}',
      ylabel_horizontal: bool = True,
      shared_y_scale: bool = False,
      alpha: float = 0.8,
      order_tdata_by_mean: bool = True,
      max_num_tracks: int = 50,
      legend_loc: str = 'upper right',
      **kwargs,
  ):
    """Initializes the `OverlaidTracks` component.

    Args:
      tdata: A dictionary mapping track names to `TrackData` objects.
      colors: An optional dictionary mapping track names to colors.
      cmap: The colormap to use if `colors` is not provided.
      track_height: The height of each track.
      ylabel_template: A template for the y-axis labels.
      ylabel_horizontal: Whether to make the y-axis labels horizontal.
      shared_y_scale: Whether to use the same y-axis scale for all tracks.
      alpha: The transparency of the tracks.
      order_tdata_by_mean: Whether to order the tracks by their mean value (in
        descending order).
      max_num_tracks: The maximum number of tracks to plot.
      legend_loc: The location of the legend (such as 'upper left' or 'best').
        See the matplotlib Axes legend documentation for more details and
        options. If None, no legend is shown.
      **kwargs: Additional keyword arguments to pass to the plotting function.

    Raises:
      ValueError: If the shapes or metadata of the track data do not match,
                  or if the number of tracks exceeds `max_num_tracks`, or if
                  the track data has more than one positional axis, or if the
                  interval is not set, or if colors are passed but do not match
                  the track data names.
    """
    self._tdata = tdata
    self._colors = colors
    self._cmap = cmap
    self._track_height = track_height
    self._ylabel_template = ylabel_template
    self._ylabel_horizontal = ylabel_horizontal
    self._shared_y_scale = shared_y_scale
    self._alpha = alpha
    self._order_tdata_by_mean = order_tdata_by_mean
    self._kwargs = kwargs
    self._first_tdata = list(tdata.values())[0]
    self._legend_loc = legend_loc

    if len(set(data.values.shape for data in tdata.values())) != 1:
      raise ValueError('Shapes of track data values must be the same.')

    if not all(
        self._first_tdata.metadata.equals(data.metadata)
        for data in tdata.values()
    ):
      raise ValueError('Metadata of track data must be the same.')

    if self._first_tdata.num_tracks > max_num_tracks:
      raise ValueError(
          f'Too many tracks to plot: {self._first_tdata.num_tracks} >'
          f' {max_num_tracks}.'
      )

    # If a colors dict is passed, then there should be a matching color for each
    # track data name.
    if self._colors is not None:
      if self._tdata.keys() != self._colors.keys():
        raise ValueError(
            f'If passing colors, each tdata name {list(self._tdata.keys())} '
            'must have an associated color.'
        )
    # Otherwise, we define a color from a cmap.
    else:
      cmap = plt.get_cmap(self._cmap)
      colors = cmap(np.linspace(0, 1, len(tdata)))
      self._colors = dict(zip(self._tdata.keys(), colors))

    if len(self._first_tdata.positional_axes) != 1:
      raise ValueError(
          'Only track_data with 1 positional axis is supported in'
          ' OverlaidTracks.'
      )
    if self._first_tdata.interval is None:
      raise ValueError('.interval needs to be set in track_data.')

    if self._shared_y_scale:
      # We compute min and max over all the track data arrays in the tdata dict.
      all_values = np.stack([arr.values for arr in self._tdata.values()])
      self._vmin = np.min(all_values)
      self._vmax = np.max(all_values)

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._track_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    return self._first_tdata.num_tracks

  def plot_ax(
      self,
      ax: matplotlib.axes.Axes,
      axis_index: int,
      interval: genome.Interval,
  ):
    """Plots the overlaid tracks on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """

    def _maybe_slice_tdata(td):
      """Slices the track data to the interval if passed."""
      # If an interval is passed, zoom in on that specific sub-interval.
      if interval is not None:
        return td.slice_by_interval(interval, match_resolution=True)
      else:
        return td

    def _make_ordered_dict_by_mean(tdata):
      """Reorders track data dict by mean (descending) for better plotting."""
      mean_tuples = [
          (name, np.mean(td.values, dtype=np.float64))
          for name, td in tdata.items()
      ]
      sorted_mean_tuples = sorted(
          mean_tuples, key=lambda item: item[1], reverse=True
      )

      # Extract names in the sorted order and return ordered tdata.
      sorted_names = [name for name, _ in sorted_mean_tuples]
      ordered_tdata = {name: tdata[name] for name in sorted_names}
      return ordered_tdata

    # Maybe slice the track data to the interval, and reorder by mean.
    tdata_sliced = {
        name: _maybe_slice_tdata(td) for name, td in self._tdata.items()
    }
    self._tdata_ordered = (
        _make_ordered_dict_by_mean(tdata_sliced)
        if self._order_tdata_by_mean
        else tdata_sliced
    )

    for name, tdata in self._tdata_ordered.items():
      assert tdata.interval is not None
      x = (
          np.arange(tdata.values.shape[0]) * tdata.resolution
          + tdata.interval.start
          + tdata.resolution / 2
      )
      arr = tdata.values[:, axis_index]

      if self._shared_y_scale:
        ax.set_ylim(self._vmin, self._vmax)

      # Plot the two tracks on the same axis. We plot the larger values first so
      # that the overlap is more visible.
      ax.plot(
          x,
          arr,
          alpha=self._alpha,
          **self._kwargs,
          c=self._colors[name],
      )
      if axis_index == 0 and self._legend_loc is not None:
        ax.legend(self._tdata_ordered.keys(), loc=self._legend_loc)

    if self._ylabel_template:
      _set_ylabel(ax, self._get_ylabel(axis_index), self._ylabel_horizontal)

  def _get_ylabel(self, axis_index: int) -> str:
    """Returns the y-axis label for the given track."""
    # Metadata equality has already been checked so here we grab the first one.
    metadata = self._tdata[list(self._tdata.keys())[0]].metadata
    row = metadata.iloc[axis_index]
    return self._ylabel_template.format(**row.to_dict())


class ContactMaps(AbstractComponent):
  """Component for visualizing contact maps.

  The `vmin` and `vmax` parameters control the color scaling of the heatmap.
  Values outside this range will be clipped to `vmin` or `vmax`.
  """

  def __init__(
      self,
      tdata: track_data.TrackData,
      track_height: float = 10.0,
      vmin: float | None = -1.0,
      vmax: float | None = 2.0,
      norm: matplotlib.colors.TwoSlopeNorm | None = None,
      ylabel_horizontal: bool = True,
      ylabel_template: str = '{name}',
      cmap: matplotlib.colors.LinearSegmentedColormap | None = None,
      max_num_tracks: int = 10,
      **kwargs,
  ):
    """Initializes the `ContactMaps` component.

    Args:
      tdata: The `TrackData` object containing the contact maps.
      track_height: The height of each contact map.
      vmin: The minimum value for the color scale.
      vmax: The maximum value for the color scale.
      norm: An optional normalization for the color scale.
      ylabel_horizontal: Whether to make the y-axis labels horizontal.
      ylabel_template: A template for the y-axis labels.
      cmap: The colormap to use for the contact maps.
      max_num_tracks: The maximum number of tracks to plot.
      **kwargs: Additional keyword arguments to pass to the plotting function.

    Raises:
      ValueError: If the number of tracks exceeds `max_num_tracks`, or if the
        track data does not have 2 positional axes, or if the contact maps are
        not square, or if the interval is not set in the track data.
    """
    if tdata.num_tracks > max_num_tracks:
      raise ValueError(
          f'Too many tracks to plot: {tdata.num_tracks} > {max_num_tracks}.'
      )
    self._tdata = tdata
    self._resolution = tdata.resolution
    self._track_height = track_height
    self._vmin = vmin
    self._vmax = vmax
    self._norm = norm
    self._ylabel_horizontal = ylabel_horizontal
    self._ylabel_template = ylabel_template
    # TODO: b/377292012 - Add orca_color_map.
    self._cmap = (
        cmap if cmap is not None else matplotlib.pyplot.get_cmap('autumn_r')
    )
    self._kwargs = kwargs
    if len(self._tdata.positional_axes) != 2:
      raise ValueError(
          'Only track_data with 2 positional axes is supported in ContactMaps.'
      )
    if self._tdata.values.shape[0] != self._tdata.values.shape[1]:
      raise ValueError('Contact maps must be square.')

    if self._tdata.interval is None:
      raise ValueError('.interval needs to be set in track_data.')

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._track_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    return self._tdata.num_tracks

  def _get_bin_positions(
      self, interval: genome.Interval, resolution: int
  ) -> np.ndarray:
    """Gets the positions of contact map bins in chromosome coordinates."""
    bin_indices = np.arange(interval.width // resolution)
    return interval.start + (resolution * bin_indices)

  def _plot_pcolormesh(
      self,
      ax: matplotlib.axes.Axes,
      x: np.ndarray,
      y: np.ndarray,
      arr: np.ndarray,
      vmin: float | None = None,
      vmax: float | None = None,
      cmap: matplotlib.colors.Colormap | None = None,
      norm: matplotlib.colors.Normalize | None = None,
      **kwargs,
  ) -> matplotlib.collections.QuadMesh:
    """Plots the contact map heatmap using `pcolormesh`.

    Note that upsampling the contact maps to single base pair resolution and
    using something like .imshow() is infeasible since the upsampling blows up
    memory.

    Args:
      ax: The matplotlib axis to plot on.
      x: The x-axis coordinates.
      y: The y-axis coordinates.
      arr: The contact map data.
      vmin: The minimum value for the color scale.
      vmax: The maximum value for the color scale.
      cmap: The colormap to use.
      norm: An optional normalization for the color scale.
      **kwargs: Additional keyword arguments to pass to `pcolormesh`.

    Returns:
      The `matplotlib.collections.QuadMesh` object representing the plot.
    """
    if not norm:  # Cannot pass both norm and vmin/vmax simultaneously.
      return ax.pcolormesh(
          x,
          y,
          arr,
          vmin=vmin,
          vmax=vmax,
          cmap=cmap,
          **kwargs,
      )
    else:
      return ax.pcolormesh(
          x,
          y,
          arr,
          norm=norm,
          **kwargs,
      )

  def plot_ax(
      self,
      ax: matplotlib.axes.Axes,
      axis_index: int,
      interval: genome.Interval,
  ):
    """Plots the contact map on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """
    tdata = self._tdata
    assert tdata.interval is not None

    # If an interval is passed, zoom in on that specific sub-interval.
    if interval is not None:
      tdata = tdata.slice_by_interval(interval, match_resolution=True)
      del interval

    arr = tdata.values[:, :, axis_index]
    x = self._get_bin_positions(tdata.interval, self._resolution)

    # We shift the bin edges by half a step since pcolormesh will misalign
    # the x-axis by half a step, since the plot values are centered within each
    # bin, rather than at the bin edges.
    half_bin_width = self._resolution // 2
    x = x + half_bin_width

    # The -1 reverse ordering ensures that contact maps are plotted with
    # the diagonal going down from left to right.
    y = np.arange(arr.shape[0])[::-1]

    if not self._vmin:
      self._vmin = np.min(arr)

    if not self._vmax:
      self._vmax = np.max(arr)

    self._plot_pcolormesh(
        ax=ax,
        x=x,
        y=y,
        arr=arr,
        vmin=self._vmin,
        vmax=self._vmax,
        cmap=self._cmap,
        norm=self._norm,
        **self._kwargs,
    )

    if self._ylabel_template:
      _set_ylabel(ax, self._get_ylabel(axis_index), self._ylabel_horizontal)

  def _get_ylabel(self, axis_index: int) -> str:
    """Returns the y-axis label for the given contact map."""
    row = self._tdata.metadata.iloc[axis_index]
    return self._ylabel_template.format(**row.to_dict())


class ContactMapsDiff(ContactMaps):
  """Component for visualizing contact map differences.

  This component visualizes the difference between two contact maps. It uses
  a diverging red-blue color map with the center white color pinned to a
  value of zero, with negative values being blue and positive values being red.

  The `vmin` and `vmax` parameters control the color scaling of the heatmap.
  Values outside this range will be clipped to `vmin` or `vmax`.
  """

  def __init__(
      self,
      tdata: track_data.TrackData,
      track_height: float = 10.0,
      vmin: float | None = -1.0,
      vmax: float | None = 1.0,
      ylabel_horizontal: bool = True,
      ylabel_template: str = '{name}',
      cmap: matplotlib.colors.LinearSegmentedColormap | str | None = 'RdBu_r',
      max_num_tracks: int = 10,
      **kwargs,
  ):
    """Initializes the `ContactMapsDiff` component.

    Args:
      tdata: The `TrackData` object containing the contact map differences.
      track_height: The height of each contact map.
      vmin: The minimum value for the color scale.
      vmax: The maximum value for the color scale.
      ylabel_horizontal: Whether to make the y-axis labels horizontal.
      ylabel_template: A template for the y-axis labels.
      cmap: The colormap to use for the contact maps.
      max_num_tracks: The maximum number of tracks to plot.
      **kwargs: Additional keyword arguments to pass to the plotting function.
    """
    self._norm = plt_colors.TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)

    super().__init__(
        tdata=tdata,
        track_height=track_height,
        vmin=vmin,
        vmax=vmax,
        ylabel_horizontal=ylabel_horizontal,
        ylabel_template=ylabel_template,
        cmap=cmap,
        max_num_tracks=max_num_tracks,
        **kwargs,
    )


def _set_ylabel(ax: matplotlib.axes.Axes, ylabel: str, horizontal: bool):
  """Sets the y-axis label.

  Args:
    ax: The matplotlib axis to set the label on.
    ylabel: The label text.
    horizontal: Whether to make the label horizontal.
  """
  if ylabel:
    if horizontal:
      ax.set_ylabel(
          ylabel,
          rotation=0,
          multialignment='center',
          va='center',
          ha='right',
          labelpad=5,
      )
    else:
      ax.set_ylabel(ylabel)


class TranscriptAnnotation(AbstractComponent):
  """Visualizes transcript annotations."""

  def __init__(
      self,
      transcripts: Sequence[transcript_utils.Transcript],
      adaptive_fig_height: bool = True,
      fig_height: float = 1.0,
      transcript_style: plot_transcripts.TranscriptStyle = (
          plot_transcripts.TranscriptStylePreset.MINIMAL.value
      ),
      plot_labels_once: bool = True,
      label_name: str = 'gene_name',
      **kwargs,
  ):
    """Initializes the `TranscriptAnnotation` component.

    Args:
      transcripts: A sequence of `Transcript` objects to visualize.
      adaptive_fig_height: Whether to adjust the figure height based on the
        number of transcripts.
      fig_height: The base figure height.
      transcript_style: The style to use for plotting transcripts. The options
        are defined in `plot_transcripts.TranscriptStylePreset`.
      plot_labels_once: Whether to plot labels only once per transcript.
      label_name: The attribute of the transcript to use for labels.
      **kwargs: Additional keyword arguments to pass to the plotting function.
    """
    self._transcripts = transcripts
    self._adaptive_fig_height = adaptive_fig_height
    self._fig_height = fig_height
    self._kwargs = kwargs
    self._kwargs['label_name'] = label_name
    self._kwargs['transcript_style'] = transcript_style
    self._kwargs['plot_labels_once'] = plot_labels_once

    self._num_transcripts = len(self._transcripts)
    # TODO(b/377291518): adaptive fig height should in theory scale with the
    # number of transcripts in the sub-interval, not the total number of
    # transcripts passed, but this is a bit tricky to implement in the code and
    # this approach seems to work well enough for now.
    if self._adaptive_fig_height:
      self._fig_height = max(0.05 * self._num_transcripts * self._fig_height, 1)

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._fig_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    return 1

  def plot_ax(
      self, ax: matplotlib.axes.Axes, axis_index: int, interval: genome.Interval
  ):
    """Plots the transcript annotations on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """
    # Update transcripts to only those overlapping interval.
    transcripts = [
        t for t in self._transcripts if t.transcript_interval.overlaps(interval)
    ]
    ax.set_yticklabels([])
    ax.set_yticks([])
    ax.spines['left'].set_visible(False)
    plot_transcripts.plot_transcripts(ax, transcripts, interval, **self._kwargs)


class SeqLogo(AbstractComponent):
  """Visualizes a sequence logo."""

  def __init__(
      self,
      scores: Float32[np.ndarray, 'S A'],
      scores_interval: genome.Interval,
      fig_height: float = 1.0,
      alphabet: str = 'ACGT',
      max_width: int = 1000,
      ylabel: str = '',
      ylabel_horizontal: bool = True,
      ylim: tuple[float, float] | None = None,
      **kwargs,
  ):
    """Initializes the `SeqLogo` component.

    Args:
      scores: A numpy array of shape (sequence_length, alphabet_size) containing
        the sequence logo scores.
      scores_interval: The genomic interval corresponding to the scores.
      fig_height: The height of the figure.
      alphabet: The alphabet used in the sequence logo.
      max_width: The maximum width of the sequence logo to plot.
      ylabel: An optional label for the y-axis.
      ylabel_horizontal: Whether to make the y-axis label horizontal.
      ylim: An optional range to set for the y-axis.
      **kwargs: Additional keyword arguments to pass to the plotting function.
    """
    self._scores = scores
    self._scores_interval = scores_interval
    self._fig_height = fig_height
    self._alphabet = alphabet
    self._max_width = max_width
    self._ylabel = ylabel
    self._ylabel_horizontal = ylabel_horizontal
    self._ylim = ylim
    self._kwargs = kwargs

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._fig_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    return 1

  def plot_ax(
      self, ax: matplotlib.axes.Axes, axis_index: int, interval: genome.Interval
  ):
    """Plots the sequence logo on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """
    intersection = self._scores_interval.intersect(interval)
    if intersection is None or intersection.width > self._max_width:
      return
    relative_start = intersection.start - self._scores_interval.start
    scores = self._scores[
        relative_start : (relative_start + intersection.width)
    ]

    plot_lib.seqlogo(
        scores,
        ax=ax,
        alphabet=self._alphabet,
        start=intersection.start,
        one_based=False,
        **self._kwargs,
    )

    _set_ylabel(ax, self._ylabel, self._ylabel_horizontal)
    if self._ylim is not None:
      ax.set_ylim(self._ylim)


class Sashimi(AbstractComponent):
  """Visualizes splice junctions as a Sashimi plot."""

  def __init__(
      self,
      junction_track: junction_data.JunctionData,
      fig_height: float = 1.0,
      filter_threshold: float | None = None,
      ylabel_template: str = '{name}',
      ylabel_horizontal: bool = True,
      annotate_counts: bool = True,
      normalize_values: bool = True,
      interval_contained: bool = True,
      rng: np.random.Generator | None = None,
  ):
    """Initializes the `Sashimi` component.

    Args:
      junction_track: A `JunctionData` object to visualize.
      fig_height: The height of the figure.
      filter_threshold: The minimum value for a junction to be included in the
        plot. This is typically based on the normalized read count. If None,
        filter out junction values below 5% of the maximum value.
      ylabel_template: A template for the y-axis labels.
      ylabel_horizontal: Whether to make the y-axis label horizontal.
      annotate_counts: Whether to annotate the junctions with read counts.
      normalize_values: Whether to normalize the values to a constant sum.
      interval_contained: Whether to only plot junctions contained in the
        interval.
      rng: Optional random number generator to use for jittering junction paths.
        If unset will use NumPy's default random number generator.
    """
    if normalize_values:
      self._junction_track = junction_track.normalize_values()
    else:
      self._junction_track = junction_track
    self._fig_height = fig_height
    self._filter_threshold = filter_threshold
    self._ylabel_template = ylabel_template
    self._ylabel_horizontal = ylabel_horizontal
    self._annotate_counts = annotate_counts
    self._interval_contained = interval_contained
    self._rng = rng or np.random.default_rng()

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._fig_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    # Metadata for JunctionData do not have strand information.
    # Strand information is in JunctionData.intervals.
    return self._junction_track.num_tracks * len(
        self._junction_track.possible_strands
    )

  def _get_strand_and_metadata_index(self, axis_index: int) -> tuple[str, int]:
    """Returns the strand and metadata index for the given axis index."""
    if len(self._junction_track.possible_strands) == 1:
      strand = self._junction_track.possible_strands[0]
      metadata_index = axis_index
    else:
      strand = '+' if axis_index % 2 == 0 else '-'
      metadata_index = axis_index // 2
    return strand, metadata_index

  def plot_ax(
      self, ax: matplotlib.axes.Axes, axis_index: int, interval: genome.Interval
  ):
    """Plots the Sashimi plot on the given axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """
    strand, metadata_index = self._get_strand_and_metadata_index(axis_index)
    track_name = self._junction_track.metadata.iloc[metadata_index]['name']
    junction_track = self._junction_track.intersect_with_interval(interval)
    junctions = junction_data.get_junctions_to_plot(
        predictions=junction_track,
        strand=strand,
        name=track_name,
        k_threshold=self._filter_threshold,
    )
    if self._interval_contained:
      junctions = [j for j in junctions if interval.contains(j)]
    else:
      junctions = [j for j in junctions if j.overlaps(interval)]

    plot_lib.sashimi_plot(
        junctions,
        ax=ax,
        interval=interval,
        filter_threshold=0,
        annotate_counts=self._annotate_counts,
        rng=self._rng,
    )
    ax.set_yticklabels([])
    ax.set_yticks([])
    ax.spines['left'].set_visible(False)
    if self._ylabel_template:
      _set_ylabel(ax, self._get_ylabel(axis_index), self._ylabel_horizontal)

  def _get_ylabel(self, axis_index: int) -> str:
    """Returns the y-axis label for the given axis index."""
    strand, metadata_index = self._get_strand_and_metadata_index(axis_index)
    row = self._junction_track.metadata.iloc[metadata_index]
    row = row.to_dict()
    row['strand'] = strand
    return self._ylabel_template.format(**row)


class EmptyComponent(AbstractComponent):
  """An empty plotting component."""

  def __init__(self, fig_height: float = 1.0):
    """Initializes the `EmptyComponent`.

    Args:
      fig_height: The height of the figure.
    """
    self._fig_height = fig_height

  def get_ax_height(self, axis_index: int) -> float:
    """Returns the height of the axis."""
    return self._fig_height

  @property
  def num_axes(self) -> int:
    """Returns the number of matplotlib axes required by the component."""
    return 1

  def plot_ax(
      self, ax: matplotlib.axes.Axes, axis_index: int, interval: genome.Interval
  ):
    """Plot an empty axis, removing all labels and spines from the axis.

    Args:
      ax: The matplotlib axis to plot on.
      axis_index: The index of the axis.
      interval: The genomic interval to plot.
    """
    ax.set_yticklabels([])
    ax.set_yticks([])
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)


class AbstractAnnotation(abc.ABC):
  """Abstract base class for plot annotations.

  Annotations are visual elements that can be added to plots to highlight
  specific features or regions. This class defines the common interface
  for all annotations.

  Attributes:
    annotations: A sequence of `Variant` or `Interval` objects representing the
      annotations.
    colors: An optional string or sequence of strings specifying the colors of
      the annotations.
    labels: An optional sequence of strings to use as labels for the
      annotations.
    use_default_labels: Whether to use default labels for the annotations if
      `labels` is not provided.
  """

  def __init__(
      self,
      annotations: Sequence[genome.Variant] | Sequence[genome.Interval],
      colors: str | Sequence[str] | None,
      labels: Sequence[str] | None,
      use_default_labels: bool,
  ):
    """Initializes the `AbstractAnnotation` class.

    Args:
      annotations: A sequence of `Variant` or `Interval` objects.
      colors: An optional string or sequence of strings specifying colors.
      labels: An optional sequence of strings to use as labels.
      use_default_labels: Whether to use default labels if `labels` is not
        provided.

    Raises:
      ValueError: If the length of `colors` or `labels` does not match the
        length of `annotations`.
    """
    self._annotations = annotations
    self._colors = colors
    self._labels = labels
    self._use_default_labels = use_default_labels

    # Pre-processing / validation of inputs.
    num_annotations = len(self._annotations)
    if (self._colors is not None) and (not isinstance(self._colors, str)):
      if len(self._colors) != num_annotations:
        raise ValueError(
            'Colors must have the same length as intervals/variants or just'
            ' a single color string.'
        )
    if self._labels is not None:
      if len(self._labels) != num_annotations:
        raise ValueError(
            'Labels must have the same length as intervals/variants.'
        )

  @abc.abstractmethod
  def plot_ax(
      self, ax: matplotlib.axes.Axes, interval: genome.Interval, hspace: float
  ):
    """Adds the annotation to an individual axis.

    Args:
      ax: The matplotlib axis to add the annotation to.
      interval: The genomic interval to plot.
      hspace: The vertical space between subplots.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def plot_labels(
      self,
      ax: matplotlib.axes.Axes,
      interval: genome.Interval,
      label_height_factor: float,
  ):
    """Adds labels for the annotation to an axis.

    Args:
      ax: The matplotlib axis to add the labels to.
      interval: The genomic interval to plot.
      label_height_factor: A scaling factor for the label height.
    """
    raise NotImplementedError

  @property
  def is_variant(self) -> bool:
    """Returns True if the annotation is a variant annotation."""
    return isinstance(self._annotations[0], genome.Variant)

  @property
  def has_labels(self) -> bool:
    """Returns True if the annotation has labels."""
    return (self._labels is not None) | (
        self.is_variant and self._use_default_labels
    )

  def add_label(
      self,
      ax: matplotlib.axes.Axes,
      label_x_position: float,
      label: str,
      angle: float,
      label_height_factor: float,
      label_position: str = 'left',
  ):
    """Adds a single angled label to an axis.

    Args:
      ax: The matplotlib axis to add the label to.
      label_x_position: The x position of the label.
      label: The label text.
      angle: The angle of the label.
      label_height_factor: A scaling factor for the label height.
      label_position: The (horizontal) placement of the label, relative to the x
        position. Can be any position string accepted by the horizontalalignment
        argument of matplotlib.axes.Axes.text.
    """
    ylims = ax.get_ylim()
    label_height = ylims[0] + label_height_factor * np.diff(ylims)[0]
    ax.text(
        label_x_position,
        label_height,
        label,
        color='black',
        fontsize=10,
        rotation=angle,
        ha=label_position,
        va='bottom',
    )
    ax.axvline(
        label_x_position, ymax=label_height * 0.95, color='black', alpha=0.1
    )


class IntervalAnnotation(AbstractAnnotation):
  """Visualizes intervals as rectangles across all plot components.

  A rectangle is drawn for each interval and overlaid on top of the final plot,
  spanning all plot components.
  """

  def __init__(
      self,
      intervals: Sequence[genome.Interval],
      colors: str | Sequence[str] = 'darkgray',
      alpha: float = 0.2,
      labels: Sequence[str] | None = None,
      use_default_labels: bool = True,
      label_angle: float = 15,
  ):
    """Initializes the `IntervalAnnotation` class.

    Args:
      intervals: A sequence of `Interval` objects to annotate.
      colors: An optional string or sequence of strings specifying the colors of
        the interval annotation.
      alpha: The transparency of the interval annotation.
      labels: An optional sequence of strings to use as labels for the
        intervals.
      use_default_labels: Whether to use default labels for the intervals if
        `labels` is not provided.
      label_angle: The angle of the interval labels.
    """
    super().__init__(intervals, colors, labels, use_default_labels)
    self._label_angle = label_angle
    self._alpha = alpha
    self._intervals = intervals

  def plot_ax(
      self,
      ax: matplotlib.axes.Axes,
      interval: genome.Interval,
      hspace: float = 0.0,
  ):
    """Adds the interval annotation to an individual axis.

    Args:
      ax: The matplotlib axis to add the annotation to.
      interval: The genomic interval to plot.
      hspace: The vertical space between subplots.
    """
    for i, interval_i in enumerate(self._intervals):
      if isinstance(self._colors, str):
        color = self._colors
      else:
        color = self._colors[i]
      # Only plot the piece of the annotation that intersects with the plotting
      # interval.
      intersection = interval_i.intersect(interval)
      if intersection is None:
        continue
      ax.axvspan(
          interval_i.start,
          interval_i.end,
          # Set y-maximum to be the top of the axis, including hspace.
          ymax=(1 + hspace) * 1.03,
          alpha=self._alpha,
          facecolor=color,
          # Set edgecolor to None for intervals to avoid horizongal lines
          # between axes, but keep it for variants. As variants are typically
          # a very thin rectangle, removing the edges results in the rectangle
          # not being visible.
          edgecolor=None,
          clip_on=False,
      )

  def plot_labels(
      self,
      ax: matplotlib.axes.Axes,
      interval: genome.Interval,
      label_height_factor: float,
  ):
    """Adds interval labels to an axis.

    Args:
      ax: The matplotlib axis to add the labels to.
      interval: The genomic interval to plot.
      label_height_factor: A scaling factor for the label height.
    """
    # Only add labels if they are provided.
    if self.has_labels:
      for i, interval_i in enumerate(self._intervals):
        label = self._labels[i]
        # Place label in the middle of the rectangle.
        # Only plot the piece of the annotation that intersects with the
        # plotting interval.
        intersection = interval_i.intersect(interval)
        if intersection is None:
          continue

        self.add_label(
            ax,
            label_x_position=np.mean((interval_i.start, interval_i.end)),
            label=label,
            angle=self._label_angle,
            label_height_factor=label_height_factor,
        )


class VariantAnnotation(AbstractAnnotation):
  """Visualizes variants as thin line-like rectangles across plot components."""

  def __init__(
      self,
      variants: Sequence[genome.Variant],
      colors: str | Sequence[str] = 'orange',
      alpha: float = 0.8,
      labels: Sequence[str] | None = None,
      use_default_labels: bool = True,
      label_angle: float = 15,
      label_position: str = 'left',
  ):
    """Initializes the `VariantAnnotation` class.

    Args:
      variants: A sequence of `Variant` objects to annotate.
      colors: An optional string or sequence of strings specifying the colors of
        the variant annotation.
      alpha: The transparency of the variant annotation.
      labels: An optional sequence of strings to use as labels for the variants.
      use_default_labels: Whether to use default labels for the variants if
        `labels` is not provided.
      label_angle: The angle of the variant labels.
      label_position: The (horizontal) placement of the variant label, relative
        to the variant position. Can be any position string accepted by the
        horizontalalignment argument of matplotlib.axes.Axes.text.
    """
    super().__init__(variants, colors, labels, use_default_labels)
    self._label_angle = label_angle
    self._alpha = alpha
    self._variants = variants
    self._label_position = label_position

  def plot_ax(
      self,
      ax: matplotlib.axes.Axes,
      interval: genome.Interval,
      hspace: float = 0.0,
  ):
    """Adds a variant annotation to an individual axis.

    Args:
      ax: The matplotlib axis to add the annotation to.
      interval: The genomic interval to plot.
      hspace: The vertical space between subplots.
    """
    for i, variant in enumerate(self._variants):
      if isinstance(self._colors, str):
        color = self._colors
      else:
        color = self._colors[i]
      interval_i = variant.reference_interval
      # Only plot the piece of the annotation that intersects with the plotting
      # interval.
      intersection = interval_i.intersect(interval)
      if intersection is None:
        continue
      ax.axvspan(
          interval_i.start,
          interval_i.end,
          # Set y-maximum to be the top of the axis, including hspace.
          ymax=(1 + hspace) * 1.03,
          alpha=self._alpha,
          facecolor=color,
          # Set edgecolor to None for intervals to avoid horizongal lines
          # between axes, but keep it for variants. As variants are typically
          # a very thin rectanle, removing the edges results in the rectangle
          # not being visible.
          edgecolor=color,
          clip_on=False,
      )

  def plot_labels(
      self,
      ax: matplotlib.axes.Axes,
      interval: genome.Interval,
      label_height_factor: float,
  ):
    """Adds variant labels to an axis.

    Args:
      ax: The matplotlib axis to add the labels to.
      interval: The genomic interval to plot.
      label_height_factor: A scaling factor for the label height.
    """
    # Only add labels if they are provided.
    if self.has_labels:
      for i, variant in enumerate(self._variants):
        interval_i = variant.reference_interval
        # Using truncated string method for genome.Variant class to get default
        # labels for variants.
        label = (
            variant.as_truncated_str(max_length=20)
            if self._use_default_labels
            else self._labels[i]
        )
        # Place label in the middle of the rectangle.
        # Only plot the piece of the annotation that intersects with the
        # plotting interval.
        intersection = interval_i.intersect(interval)
        if intersection is None:
          continue

        self.add_label(
            ax,
            label_x_position=np.mean((interval_i.start, interval_i.end)),
            label=label,
            angle=self._label_angle,
            label_height_factor=label_height_factor,
            label_position=self._label_position,
        )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import ontology


class OntologyTest(parameterized.TestCase):

  def test_ontology_term_by_curie(self):
    term = ontology.from_curie('UBERON:0000005')
    self.assertEqual(term.ontology_curie, 'UBERON:0000005')
    self.assertEqual(term.type, ontology.OntologyType.UBERON)
    self.assertEqual(term.id, 5)

  @parameterized.parameters(
      ('UBERON:0000005',),
      ('CL:0000005',),
      ('EFO:0000005',),
      ('NTR:0000512',),
  )
  def test_proto_roundtrip(self, curie: str):
    term = ontology.from_curie(curie)
    proto = term.to_proto()
    round_trip = ontology.from_proto(proto)
    self.assertEqual(term, round_trip)

  def test_ontology_terms_from_curies(self):
    terms = ontology.from_curies(['UBERON:0000005', 'CL:0000005'])
    self.assertEqual(
        terms,
        [
            ontology.OntologyTerm(ontology.OntologyType.UBERON, 5),
            ontology.OntologyTerm(ontology.OntologyType.CL, 5),
        ],
    )

  @parameterized.named_parameters(
      ('invalid_curie', 'invalid:0000005', 'Cannot parse ontology_type'),
      ('lowercase_curie', 'uberon:0000005', 'Cannot parse ontology_type'),
      ('too_many_colons', 'UBERON:0000005:0000005', 'Invalid ontology_curie'),
      ('missing_colon', 'foo_bar', "Invalid ontology_curie='foo_bar'"),
      (
          'invalid_id',
          'UBERON:foo',
          r"invalid literal for int\(\) with base 10: 'foo'",
      ),
  )
  def test_invalid_curie_raises_error(self, curie: str, expected_error: str):
    with self.assertRaisesRegex(ValueError, expected_error):
      ontology.from_curie(curie)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Track data container analogous to AnnData."""

from collections.abc import Sequence
import copy
import dataclasses
import enum
from typing import Any, Union

from alphagenome import typing
from alphagenome.data import genome
from alphagenome.data import ontology
from jaxtyping import Bool, Float32, Int32  # pylint: disable=g-multiple-import, g-importing-member
import numpy as np
import pandas as pd

# Required columns: name, strand.
# Optional standardized columns: cell_type, assay, padding.
TrackMetadata = pd.DataFrame
PositionalIndex = slice | genome.Interval | int
TrackIndex = np.ndarray | Sequence[int] | Sequence[str] | slice | int | str
Index = PositionalIndex | tuple[PositionalIndex, TrackIndex] | TrackIndex


@enum.unique
class AggregationType(enum.Enum):
  """Aggregation types for downsampling/upsampling track resolutions.

  SUM: Sum pooling, where values within a bin are summed. This is recommended
    for continuous tracks where the total value within a bin is meaningful (e.g.
    read counts, coverage).
  MAX: Max pooling, where the maximum value within a bin is selected. This is
    recommended for binary tracks (e.g., gene masks, regions of interest).
  """

  SUM = 'sum'
  MAX = 'max'


@typing.jaxtyped
@dataclasses.dataclass(frozen=True)
class TrackData:
  """Container for storing track values and metadata.

  `TrackData` stores multiple genomic tracks at the same resolution, stacked
  into an ND matrix of shape (positional_bins, num_tracks). It also contains
  metadata information as a pandas DataFrame with `num_tracks` rows.

  Metadata DataFrame has two main required columns:

    * name: The name of the track.
    * strand: The strand of the track ('+', '-', or '.').

  Other columns are optional.

  Valid shapes of `TrackData.values` are:

    * [positional_bins]
    * [positional_bins, num_tracks]
    * [positional_bins, positional_bins, num_tracks]
    * ...

  `TrackData` can store both model predictions and raw data. It can
  optionally hold information about the `genome.Interval` from which the data
  were derived and `.uns` for storing additional unstructured data.

  In addition to being a container, `TrackData` provides functionality for
  common aggregation and slicing operations.

  Attributes:
    values: A numpy array of floats or integers representing the track values.
      Positional axes have the same length. Example valid shapes are:
      [num_tracks], [positional_bins, num_tracks], and [positional_bins,
      positional_bins, num_tracks].
    metadata: A pandas DataFrame containing metadata for each track. The
      DataFrame must have at least two columns: 'name' and 'strand'.
    resolution: The resolution of the track data in base pairs.
    interval: An optional `Interval` object representing the genomic region.
    uns: An optional dictionary to store additional unstructured data.

  Raises:
    ValueError: If the number of tracks in `values` does not match the  number
      of rows in `metadata`, or if `metadata` contains duplicate (name, strand)
      pairs, or if the positional axes have different lengths, or if the
      interval width does not match the expected width.
  """

  # Use Union due to https://github.com/patrick-kidger/jaxtyping/issues/73.
  values: Union[
      Float32[np.ndarray, '*positional_bins num_tracks'],
      Int32[np.ndarray, '*positional_bins num_tracks'],
      Bool[np.ndarray, '*positional_bins num_tracks'],
  ]
  metadata: TrackMetadata
  resolution: int = 1
  interval: genome.Interval | None = None
  uns: dict[str, Any] | None = (
      # Unstructured data dict, analagous to anndata.AnnData.uns.
      None
  )

  def __post_init__(self):
    """Validates the consistency of the data."""
    if self.values.shape[-1] != len(self.metadata):
      raise ValueError(
          f'value number of tracks {self.values.shape[-1]} and '
          f'metadata {len(self.metadata)} do not match.'
      )

    if self.positional_axes:
      if len(set(np.array(self.values.shape)[self.positional_axes])) != 1:
        raise ValueError('All positional axes must have the same length.')

      if self.interval and self.interval.width != self.width:
        raise ValueError(
            f'Interval width must match expected width. {self.interval.width=},'
            f' {self.width=}'
        )

    if not {'name', 'strand'}.issubset(self.metadata.columns):
      raise ValueError('Metadata must contain columns "name" and "strand".')

    if self.metadata[['name', 'strand']].duplicated().any():
      raise ValueError(
          'Metadata contain duplicated values for (name, strand) tuples.'
      )

  @property
  def positional_axes(self) -> list[int]:
    """Returns a list of the positional axes."""
    return list(range(self.values.ndim - 1))

  @property
  def num_tracks(self) -> int:
    """Returns the number of tracks."""
    return self.values.shape[-1]

  @property
  def width(self) -> int:
    """Returns the interval width covered by the tracks."""
    if self.positional_axes:
      return self.values.shape[0] * self.resolution
    else:
      return 0

  @property
  def names(self) -> np.ndarray:
    """Returns an array of track names (not necessarily unique)."""
    return self.metadata['name'].values

  @property
  def strands(self) -> np.ndarray:
    """Returns an array of track strands."""
    return self.metadata['strand'].values

  @property
  def ontology_terms(self) -> Sequence[ontology.OntologyTerm | None] | None:
    """Returns a list of ontology terms (if available)."""
    if 'ontology_curie' in self.metadata.columns:
      return [
          ontology.from_curie(curie) if curie is not None else None
          for curie in self.metadata['ontology_curie'].values
      ]
    else:
      return None

  def copy(self) -> 'TrackData':
    """Returns a deep copy of the `TrackData` object."""
    if self.interval:
      interval = self.interval.copy()
    else:
      interval = None
    return TrackData(
        self.values.copy(),
        resolution=self.resolution,
        metadata=self.metadata.copy(),
        interval=interval,
        uns=copy.deepcopy(self.uns),
    )

  def bin_index(self, relative_position: int) -> int:
    """Returns the bin index for a relative position.

    Args:
      relative_position: The relative position within the interval.

    Returns:
      The corresponding bin index.
    """
    return relative_position // self.resolution

  def slice_by_positions(self, start: int, end: int) -> 'TrackData':
    """Slices the track data along the positional axes.

    The slicing follows Python slicing conventions (0 indexed, and includes
    elements up to end-1).

    Args:
      start: The 1-bp resolution start position for slicing.
      end: The 1-bp resolution end position for slicing.

    Returns:
      A new `TrackData` object with the sliced values.

    Raises:
      ValueError: If (end - start) is greater than the width, or if (end -
      start) is not divisible by the resolution.
    """
    if (end - start) > self.width:
      raise ValueError(
          'When slicing track data, (end - start) must be less than or '
          'equal to width.'
      )

    if (end - start) % self.resolution != 0:
      raise ValueError(
          f'end - start needs to be to be divisible by {self.resolution=}'
      )

    sl = slice(self.bin_index(start), self.bin_index(end))
    slice_list = [slice(None)] * self.values.ndim
    for i in self.positional_axes:
      slice_list[i] = sl

    interval = self.interval
    if interval:
      interval = genome.Interval(
          interval.chromosome,
          interval.start + start,
          interval.start + end,
          strand=interval.strand,
          name=interval.name,
          info=interval.info,
      )

    return TrackData(
        self.values[tuple(slice_list)],
        resolution=self.resolution,
        metadata=self.metadata,
        interval=interval,
        uns=self.uns,
    )

  def slice_by_interval(
      self, interval: genome.Interval, match_resolution: bool = False
  ) -> 'TrackData':
    """Slices the track data using a `genome.Interval`.

    Args:
      interval: The interval to slice to.
      match_resolution: If True, the interval will first be extended to make
        sure the width is divisible by resolution.

    Returns:
      A new `TrackData` object sliced to the interval.

    Raises:
      ValueError: If `.interval` is not specified or if the specified interval
        is not fully contained within the current interval.
    """
    if self.interval is None:
      raise ValueError(
          '.interval is needs to be specified for slice_by_interval.'
      )
    if not self.interval.contains(interval):
      raise ValueError(
          f'Interval {self.interval=} does not fully contain {interval=}.'
      )
    start = interval.start - self.interval.start
    end = interval.end - self.interval.start

    if match_resolution and self.resolution != 1:
      start = int(np.floor(start / self.resolution) * self.resolution)
      end = int(np.ceil(end / self.resolution) * self.resolution)
    return self.slice_by_positions(start, end)

  def pad(self, start_pad: int, end_pad: int) -> 'TrackData':
    """Pads the track data along positional axes.

    Args:
      start_pad: The amount of padding to add at the beginning.
      end_pad: The amount of padding to add at the end.

    Returns:
      A new `TrackData` object with padded values.

    Raises:
      ValueError: If `start_pad` or `end_pad` is not divisible by the
      resolution.
    """
    if start_pad == 0 and end_pad == 0:
      return self
    if start_pad % self.resolution != 0:
      raise ValueError(f'start_pad needs to be divisible by {self.resolution}')
    if end_pad % self.resolution != 0:
      raise ValueError(f'end_pad needs to be divisible by {self.resolution}')

    pad = [(0, 0)] * self.values.ndim
    for axis in self.positional_axes:
      pad[axis] = (start_pad // self.resolution, end_pad // self.resolution)

    return TrackData(
        np.pad(self.values, tuple(pad)),
        resolution=self.resolution,
        metadata=self.metadata,
        interval=None,  # Padding invalidates the interval.
        uns=self.uns,
    )

  def resize(self, width: int) -> 'TrackData':
    """Resizes the track data by cropping or padding with a fixed center.

    Args:
      width: The desired width in base pairs.

    Returns:
      A new `TrackData` object with resized values.

    Raises:
      ValueError: If `width` is not divisible by the resolution.
    """
    if width == self.width:
      return self
    elif width > self.width:
      if width % self.resolution != 0:
        raise ValueError(f'width needs to be divisible by {self.resolution}')
      pad_amount = (width - self.width) // self.resolution
      pad_start = (pad_amount // 2 + pad_amount % 2) * self.resolution
      pad_end = (pad_amount // 2) * self.resolution
      return self.pad(pad_start, pad_end)
    else:
      crop_amount = (self.width - width) // self.resolution
      start = (crop_amount // 2 + crop_amount % 2) * self.resolution
      return self.slice_by_positions(start, start + width)

  def upsample(
      self,
      resolution: int,
      aggregation_type: AggregationType = AggregationType.SUM,
  ) -> 'TrackData':
    """Upsamples the track data to a higher resolution by repeating existing values.

    Args:
      resolution: The desired resolution in base pairs.
      aggregation_type: The aggregation method to use for pooling the values.

    Returns:
      A new `TrackData` object with upsampled values.

    Raises:
      ValueError: If `resolution` is not lower than the current resolution
        or not divisible by the current resolution.
    """
    if resolution == self.resolution:
      return self
    if resolution > self.resolution:
      raise ValueError(f'Resolution must be lower than {self.resolution}')
    repeat = self.resolution // resolution
    if self.resolution % resolution != 0:
      raise ValueError(f'Resolution not divisible by {resolution}')

    values = self.values
    for axis in self.positional_axes:
      values = np.repeat(values, repeat, axis=axis)
      match aggregation_type:
        case AggregationType.SUM:
          values = values / repeat
        case AggregationType.MAX:
          pass
    return TrackData(
        values,
        resolution=resolution,
        metadata=self.metadata,
        interval=self.interval,
        uns=self.uns,
    )

  def downsample(
      self,
      resolution: int,
      aggregation_type: AggregationType = AggregationType.SUM,
  ) -> 'TrackData':
    """Downsamples the track data to a lower resolution using sum pooling.

    Args:
      resolution: The desired resolution in base pairs.
      aggregation_type: The aggregation method to use for pooling the values.

    Returns:
      A new `TrackData` object with downsampled values.

    Raises:
      ValueError: If `resolution` is not greater than the current resolution
        or not divisible by the current resolution.
    """
    if resolution == self.resolution:
      return self
    if resolution < self.resolution:
      raise ValueError(f'Resolution must be greater than {self.resolution}')
    if resolution % self.resolution != 0:
      raise ValueError(f'Resolution not divisible by {resolution}')
    pool_width = resolution // self.resolution

    values = self.values
    for axis in self.positional_axes:
      # Bring axis of interest to the front, reshape and aggregate, and reswap
      values = np.swapaxes(values, 0, axis)
      shape = list(values.shape)
      reshaped_values = values.reshape(
          [shape[0] // pool_width, pool_width] + shape[1:]
      )
      match aggregation_type:
        case AggregationType.SUM:
          values = reshaped_values.sum(axis=1)
        case AggregationType.MAX:
          values = reshaped_values.max(axis=1)
      values = np.swapaxes(values, 0, axis)

    return TrackData(
        values,
        resolution=resolution,
        metadata=self.metadata,
        interval=self.interval,
        uns=self.uns,
    )

  def change_resolution(
      self,
      resolution: int,
      aggregation_type: AggregationType = AggregationType.SUM,
  ) -> 'TrackData':
    """Changes the resolution of the track data.

    Args:
      resolution: The desired resolution in base pairs.
      aggregation_type: The aggregation method to use for pooling the values.

    Returns:
      A new `TrackData` object with the new resolution.
    """
    if resolution >= self.resolution:
      return self.downsample(resolution, aggregation_type)
    else:
      return self.upsample(resolution, aggregation_type)

  def filter_tracks(self, mask: np.ndarray | list[bool]) -> 'TrackData':
    """Filters tracks by a boolean mask.

    Args:
      mask: A boolean mask to select tracks.

    Returns:
      A new `TrackData` object with the filtered tracks.
    """
    return TrackData(
        self.values[..., mask],
        resolution=self.resolution,
        metadata=self.metadata.iloc[mask],
        interval=self.interval,
        uns=self.uns,
    )

  def filter_to_positive_strand(self) -> 'TrackData':
    """Filters tracks to the positive DNA strand."""
    return self.filter_tracks(self.strands == genome.STRAND_POSITIVE)

  def filter_to_negative_strand(self) -> 'TrackData':
    """Filters tracks to the negative DNA strand."""
    return self.filter_tracks(self.strands == genome.STRAND_NEGATIVE)

  def filter_to_nonnegative_strand(self) -> 'TrackData':
    """Filters tracks to the non-negative DNA strands (positive and unstranded)."""
    return self.filter_tracks(self.strands != genome.STRAND_NEGATIVE)

  def filter_to_nonpositive_strand(self) -> 'TrackData':
    """Filters tracks to the non-positive DNA strands (negative and unstranded)."""
    return self.filter_tracks(self.strands != genome.STRAND_POSITIVE)

  def filter_to_stranded(self) -> 'TrackData':
    """Filters tracks to stranded tracks (excluding unstranded)."""
    return self.filter_tracks(self.strands != genome.STRAND_UNSTRANDED)

  def filter_to_unstranded(self) -> 'TrackData':
    """Filters tracks to unstranded tracks."""
    return self.filter_tracks(self.strands == genome.STRAND_UNSTRANDED)

  def select_tracks_by_index(
      self, idx: np.ndarray | Sequence[int]
  ) -> 'TrackData':
    """Selects tracks by numerical index.

    Args:
      idx: A list or array of numerical indices to select tracks.

    Returns:
      A new `TrackData` object with the selected tracks.
    """
    return TrackData(
        self.values[..., idx],
        resolution=self.resolution,
        metadata=self.metadata.iloc[idx],
        interval=self.interval,
        uns=self.uns,
    )

  def select_tracks_by_name(
      self, names: np.ndarray | Sequence[str]
  ) -> 'TrackData':
    """Selects tracks by name.

    Args:
      names: A list or array of track names to select.

    Returns:
      A new `TrackData` object with the selected tracks.
    """
    track_idx = pd.Series(np.arange(self.num_tracks), index=self.names)
    return self.select_tracks_by_index(track_idx.loc[names].values)

  def __getitem__(self, index: Index) -> 'TrackData':
    """Retrieves a subset of TrackData using positional and/or track indices.

    This method allows slicing `TrackData` similar to numpy arrays or pandas
    DataFrames. The index can be a single value or a tuple.

    Args:
      index: A single index or a tuple of indices. If a single index, it's
        treated as a positional index if the `TrackData` has positional axes,
        otherwise as a track index. If a tuple, the first element specifies the
        positional slice, and the second element specifies the track index.
        Positional indices can be `int`, `slice`, or `genome.Interval`, and this
        slice is applied to all positional axes of the `values` array. Track
        indices can be `int`, `str` (track name), `slice`, `Sequence[int]`, or
        `Sequence[str]` (track names).

    Returns:
      A new `TrackData` object containing the selected subset.

    Raises:
      IndexError: If a slice step is not 1 for positional indexing.
      IndexError: If an unsupported index type is provided.
    """
    if isinstance(index, tuple):
      position_index, track_index = index
    elif self.positional_axes:
      position_index, track_index = index, None
    else:
      position_index, track_index = None, index
      if isinstance(track_index, genome.Interval):
        raise IndexError(
            'Track indexing by interval is supported only when there are'
            ' positional axes.'
        )

    tdata = self
    match position_index:
      case None:
        pass
      case int():
        tdata = tdata.slice_by_positions(position_index, position_index + 1)
      case slice():
        if position_index.step is not None and position_index.step != 1:
          raise IndexError('Slice step must be 1 for positional indexing.')
        if position_index != slice(None):
          tdata = tdata.slice_by_positions(
              position_index.start, position_index.stop
          )
      case genome.Interval():
        tdata = tdata.slice_by_interval(position_index)
      case _:
        raise IndexError(
            f'Unsupported positional index type: {type(position_index)}'
        )

    match track_index:
      case None:
        pass
      case str():
        tdata = tdata.select_tracks_by_name([track_index])
      case int():
        tdata = tdata.select_tracks_by_index([track_index])
      case slice():
        if track_index != slice(None):
          indices = np.arange(tdata.num_tracks)[track_index]
          tdata = tdata.select_tracks_by_index(indices)
      case np.ndarray() if np.issubdtype(track_index.dtype, np.character):
        tdata = tdata.select_tracks_by_name(track_index)
      case np.ndarray():
        tdata = tdata.select_tracks_by_index(track_index)
      case Sequence():
        track_index_arr = np.asarray(track_index)
        if np.issubdtype(track_index_arr.dtype, np.character):
          tdata = tdata.select_tracks_by_name(track_index_arr)
        else:
          tdata = tdata.select_tracks_by_index(track_index_arr)
      case _:
        raise IndexError(f'Unsupported track index type: {type(track_index)}')
    return tdata

  def groupby(self, column: str) -> dict[str, 'TrackData']:
    """Splits tracks into groups based on a metadata column.

    This method splits the tracks in the `TrackData` object into separate
    `TrackData` objects based on the unique values in the specified metadata
    column. It returns a dictionary where the keys are the  unique values in
    the column, and the values are new `TrackData` objects containing the
    tracks corresponding to each key.

    Args:
      column: The name of the metadata column to split by.

    Returns:
      A dictionary mapping unique values in the column to `TrackData` objects
      containing the corresponding tracks.
    """
    output = {}
    for key in self.metadata[column].unique():
      mask = (self.metadata[column] == key).values
      output[key] = self.filter_tracks(mask)
    return output

  def _reverse_complement_idx(self) -> np.ndarray:
    """Gets indices for reverse complementing the tracks.

    Returns:
      An array of indices that reorders the tracks to achieve reverse
      complementation.

    Raises:
      ValueError: If not all stranded tracks have both '+' and '-' strands,
         or if the number of '+' and '-' stranded tracks is not equal.
    """
    df_strands = pd.DataFrame({
        'name': self.names,
        'strand': self.strands,
        'old_idx': np.arange(self.num_tracks),
    })
    df_strands = df_strands[df_strands.strand != genome.STRAND_UNSTRANDED]
    df_strands.sort_values(['strand', 'name'], inplace=True)
    if np.all(df_strands.groupby('name').size() != 2):
      raise ValueError('Not all stranded tracks have both + and - strand.')
    if (df_strands.strand == genome.STRAND_POSITIVE).sum() != (
        df_strands.strand == genome.STRAND_NEGATIVE
    ).sum():
      raise ValueError(
          'We need to have the exact same number of + and - stranded tracks'
      )
    new_idx = df_strands.old_idx.values.reshape((2, -1))[::-1].ravel()
    # Swap strands by idx.
    idx = np.arange(self.num_tracks)
    idx[df_strands.old_idx.values] = new_idx
    return idx

  def reverse_complement(self) -> 'TrackData':
    """Reverse complements the track data and interval if present.

    Returns:
      A new `TrackData` object with reverse complemented tracks.
    """
    if self.interval:
      # Note that the interval needs to be stranded in order to perform
      # this operation.
      interval = self.interval.swap_strand()
    else:
      interval = None

    idx = self._reverse_complement_idx()
    slices = [slice(None)] * self.values.ndim
    slices[-1] = idx
    for axis in self.positional_axes:
      slices[axis] = slice(None, None, -1)

    return TrackData(
        self.values[tuple(slices)],
        resolution=self.resolution,
        metadata=self.metadata.iloc[idx],
        interval=interval,
        uns=self.uns,
    )

  def _check_track_data_compatibility(self, other: 'TrackData') -> None:
    """Checks if two `TrackData` objects are compatible for sum/diff.

    Args:
      other: The other `TrackData` object to compare.

    Raises:
      TypeError: If `other` is not a `TrackData` object.
      ValueError: If the intervals, resolutions, shapes, or metadata
        shapes don't match between the two objects.
    """
    if not isinstance(other, TrackData):
      raise TypeError(
          f'Unsupported type "{type(other)}". Must be a TrackData object'
      )
    if self.interval != other.interval:
      raise ValueError('Intervals must match for the two TrackData objects.')
    if self.resolution != other.resolution:
      raise ValueError('Resolutions must match for the two TrackData objects.')
    if self.values.shape != other.values.shape:
      raise ValueError('Shapes must match for the two TrackData objects.')
    if self.metadata.shape != other.metadata.shape:
      raise ValueError(
          'Metadata shapes must match for the two TrackData objects.'
      )

  def __add__(self, other: 'TrackData') -> 'TrackData':
    """Adds the values of two `TrackData` objects.

    Args:
      other: The `TrackData` object to add.

    Returns:
      A new `TrackData` object with the summed values.

    Raises:
      ValueError: If the objects are not compatible (see
        `_check_track_data_compatibility`).
      TypeError: If `other` is not a `TrackData` object.
    """
    self._check_track_data_compatibility(other)
    new_values = self.values + other.values
    return TrackData(
        values=new_values,
        metadata=self.metadata,
        resolution=self.resolution,
        interval=self.interval,
        uns=self.uns,
    )

  def __sub__(self, other: 'TrackData') -> 'TrackData':
    """Subtracts the values of two `TrackData` objects.

    Args:
      other: The `TrackData` object to subtract.

    Returns:
      A new `TrackData` object with the difference of the values.

    Raises:
      ValueError: If the objects are not compatible (see
        `_check_track_data_compatibility`).
      TypeError: If `other` is not a `TrackData` object.
    """
    self._check_track_data_compatibility(other)
    new_values = self.values - other.values
    return TrackData(
        values=new_values,
        metadata=self.metadata,
        resolution=self.resolution,
        interval=self.interval,
        uns=self.uns,
    )


def concat(
    track_datas: Sequence[TrackData],
    extra_metadata_name_and_keys: (
        tuple[str, Sequence[str | int | float]] | None
    ) = None,
) -> TrackData:
  """Concatenates multiple `TrackData` objects along the track dimension.

  This function combines multiple `TrackData` objects into a single object
  by concatenating their values and metadata. The resulting `TrackData`
  object will have the same resolution and interval as the input objects.

  Args:
    track_datas: A sequence of `TrackData` objects to concatenate. All objects
      must have the same resolution, interval, and width.
    extra_metadata_name_and_keys: An optional tuple specifying a new metadata
      column to add. The first element is the column name, and the second is a
      sequence of values to populate the column.

  Returns:
    A new `TrackData` object containing the concatenated data.

  Raises:
    ValueError: If the input `TrackData` objects have different resolutions,
      intervals, or widths, or if the length of
      `extra_metadata_name_and_keys[1]` does not match the length of
      `track_datas`.
  """
  if len(set(x.resolution for x in track_datas)) != 1:
    raise ValueError('Track data contain multiple resolutions')
  if len(set(str(x.interval) for x in track_datas)) != 1:
    raise ValueError('Track data contain multiple intervals')
  if len(set(x.width for x in track_datas)) != 1:
    raise ValueError('Track data are of different width')
  if extra_metadata_name_and_keys:
    if len(extra_metadata_name_and_keys[1]) != len(track_datas):
      raise ValueError(
          'Second element of new_metadata_name_and_keys must be the same'
          ' length as track_datas'
      )

  concatenated_metadata = (
      pd.concat(
          [x.metadata for x in track_datas],
          keys=extra_metadata_name_and_keys[1]
          if extra_metadata_name_and_keys
          else None,
          names=[extra_metadata_name_and_keys[0]]
          if extra_metadata_name_and_keys
          else None,
      )
      .reset_index(level=0, drop=extra_metadata_name_and_keys is None)
      .reset_index(drop=True)
  )
  return TrackData(
      np.concatenate(
          [x.values for x in track_datas], axis=track_datas[0].values.ndim - 1
      ),
      resolution=track_datas[0].resolution,
      metadata=concatenated_metadata,
      interval=track_datas[0].interval,
      uns=None,
  )


def interleave(
    track_datas: Sequence[TrackData], name_prefixes: Sequence[str]
) -> TrackData:
  """Interleaves multiple `TrackData` objects by alternating rows.

  This function combines multiple `TrackData` objects into a single object
  by interleaving their rows and metadata. This interleaves operation alternates
  between the trackdatas, like shuffling cards, i.e., 'abcd' interleaved with
  'efgh' would be "aebfcgdh". The resulting `TrackData` object will have the
  same resolution and interval as the input objects, but the number of tracks
  will be the sum of the tracks in the input objects.

  Args:
    track_datas: A sequence of `TrackData` objects to interleave. All objects
      must have the same shape, resolution, and interval. The order in this list
      will determine the interleaving order.
    name_prefixes: A sequence of name prefixes to add to the track names in the
      metadata to ensure uniqueness of (name, strand) pairs.

  Returns:
    A new `TrackData` object containing the interleaved data.

  Raises:
    ValueError: If the input `TrackData` objects have different shapes,
      resolutions, or intervals.
  """
  # Checks on the track data.
  shapes = set(data.values.shape for data in track_datas)
  if len(shapes) != 1:
    raise ValueError(
        'Cannot interleave track data which have different shapes. '
        f'Detected shapes: {shapes}'
    )

  if any(data.resolution != track_datas[0].resolution for data in track_datas):
    raise ValueError(
        'Cannot interleave track data which have different resolutions. '
        f'Detected shapes: {shapes}'
    )

  if any(data.interval != track_datas[0].interval for data in track_datas):
    raise ValueError(
        'Cannot interleave track data which have different intervals. '
    )

  # Interleave arrays.
  shape = list(track_datas[0].values.shape)
  shape[-1] = shape[-1] * len(track_datas)
  interleaved_data = np.empty(tuple(shape), dtype=track_datas[0].values.dtype)

  for i, data in enumerate(track_datas):
    interleaved_data[..., i :: len(track_datas)] = data.values

  # Interleave metadata.
  metadatas = []
  for prefix, data in zip(name_prefixes, track_datas):
    metadata = data.metadata.copy()
    metadata['name'] = prefix + metadata['name']
    metadatas.append(metadata)

  # Assign a new index idx that, when sorted, will produce an interleave.
  interleaved_metadata = (
      pd.concat([
          metadata.assign(
              idx=np.arange(
                  stop=(len(metadata) * len(track_datas)),
                  step=len(track_datas),
              )
              + i
          )
          for i, metadata in enumerate(metadatas)
      ])
      .sort_values('idx')
      .drop('idx', axis=1)
      .reset_index(drop=True)
  )

  return TrackData(
      values=interleaved_data,
      metadata=interleaved_metadata,
      resolution=track_datas[0].resolution,
      interval=track_datas[0].interval,
      uns={'num_interleaved_trackdatas': len(track_datas)},
  )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Client implementation for interacting with a DNA model server."""

from collections.abc import Container, Iterable, Iterator, Mapping, Sequence
import concurrent.futures
import enum
import functools
import random
import time
from typing import TypeVar

from alphagenome import tensor_utils
from alphagenome.data import genome
from alphagenome.data import junction_data
from alphagenome.data import ontology
from alphagenome.data import track_data
from alphagenome.models import dna_model
from alphagenome.models import dna_output
from alphagenome.models import interval_scorers as interval_scorers_lib
from alphagenome.models import junction_data_utils
from alphagenome.models import track_data_utils
from alphagenome.models import variant_scorers as variant_scorers_lib
from alphagenome.protos import dna_model_pb2
from alphagenome.protos import dna_model_service_pb2
from alphagenome.protos import dna_model_service_pb2_grpc
from alphagenome.protos import tensor_pb2
import anndata
import grpc
import numpy as np
import pandas as pd
import tqdm.auto


# Supported DNA sequence lengths.
SEQUENCE_LENGTH_2KB = 2**11  # 2_048
SEQUENCE_LENGTH_16KB = 2**14  # 16_384
SEQUENCE_LENGTH_100KB = 2**17  # 131_072
SEQUENCE_LENGTH_500KB = 2**19  # 524_288
SEQUENCE_LENGTH_1MB = 2**20  # 1_048_576

SUPPORTED_SEQUENCE_LENGTHS = {
    'SEQUENCE_LENGTH_2KB': SEQUENCE_LENGTH_2KB,
    'SEQUENCE_LENGTH_16KB': SEQUENCE_LENGTH_16KB,
    'SEQUENCE_LENGTH_100KB': SEQUENCE_LENGTH_100KB,
    'SEQUENCE_LENGTH_500KB': SEQUENCE_LENGTH_500KB,
    'SEQUENCE_LENGTH_1MB': SEQUENCE_LENGTH_1MB,
}

# Maximum width of an in silico mutagenesis (ISM) interval request.
# This controls the behavior of `score_ism_variants` in the client.
# When a user inputs an ISM interval that is wider than this value,
# the client will automatically split the interval into chunks of this width.
MAX_ISM_INTERVAL_WIDTH = 10

# Maximum number of variant scorers per request.
MAX_VARIANT_SCORERS_PER_REQUEST = 20

_VALID_SEQUENCE_CHARACTERS = frozenset('ACGTN')

RetryableFunction = TypeVar('RetryableFunction')


def retry_rpc(
    function: RetryableFunction,
    *,
    max_attempts: int = 5,
    initial_backoff: float = 1.25,
    backoff_multiplier: float = 1.5,
    retry_status_codes: Container[grpc.StatusCode] = frozenset(
        [grpc.StatusCode.RESOURCE_EXHAUSTED, grpc.StatusCode.UNAVAILABLE]
    ),
    jitter: float = 0.2,
) -> RetryableFunction:
  """Decorator that retries when an RPC fails.

  gRPC currently doesn't support retries for streaming RPCs. This decorator is
  an implementation of the retry logic from https://grpc.io/docs/guides/retry.

  Arguments:
    function: Callable to retry.
    max_attempts: Maximum number of attempts to make.
    initial_backoff: Initial backoff time in seconds.
    backoff_multiplier: Backoff multiplier to apply on each retry.
    retry_status_codes: Set of status codes to retry on.
    jitter: Jitter to apply to the backoff time. For example, a jitter of 0.2
      and an initial backoff of 1.5 will result in a backoff time between 1.2
      and 1.8 seconds.

  Returns:
    A decorator that retries the function on retryable RPC errors.
  """
  if backoff_multiplier < 1.0:
    raise ValueError('Backoff multiplier must be >= 1.0.')

  @functools.wraps(function)
  def wrapper(*args, **kwargs):
    attempt = 0
    current_backoff = initial_backoff + (
        random.uniform(-jitter, jitter) * initial_backoff
    )
    while True:
      try:
        attempt += 1
        return function(*args, **kwargs)
      except grpc.RpcError as e:
        error_code = e.code()  # pytype: disable=attribute-error
        if error_code not in retry_status_codes or attempt >= max_attempts:
          raise e
        time.sleep(current_backoff)
        current_backoff *= backoff_multiplier

  return wrapper


@enum.unique
class ModelVersion(enum.Enum):
  """Enumeration of all available model versions.

  A fold is a part of the genome that is held out from training, and is used for
  model validation.

  Folds are numbered from 0 to 3, and represent disjoint subsets of the genome.

  A model version that is designated FOLD_#, where # is an integer from 0 to 3,
  indicates that the model was trained with that particular fold held out.

  The ALL_FOLDS version refers to the distilled model that was trained on
  an ensemble of teacher models trained on all folds.
  """

  ALL_FOLDS = enum.auto()  # Default model version if None explicitly set.
  FOLD_0 = enum.auto()
  FOLD_1 = enum.auto()
  FOLD_2 = enum.auto()
  FOLD_3 = enum.auto()


Output = dna_output.Output
OutputMetadata = dna_output.OutputMetadata
OutputType = dna_output.OutputType
VariantOutput = dna_output.VariantOutput
Organism = dna_model.Organism

PredictResponse = TypeVar(
    'PredictResponse',
    dna_model_service_pb2.PredictSequenceResponse,
    dna_model_service_pb2.PredictIntervalResponse,
)


def _read_tensor_chunks(
    responses: Iterator[
        PredictResponse
        | dna_model_service_pb2.PredictVariantResponse
        | dna_model_service_pb2.ScoreVariantResponse
        | dna_model_service_pb2.ScoreIntervalResponse
    ],
    chunk_count: int,
) -> Iterable[tensor_pb2.TensorChunk]:
  """Helper to read a sequence of tensor chunks from a response iterator."""
  for _ in range(chunk_count):
    try:
      response = next(responses)
    except StopIteration as e:
      raise ValueError('Expected tensor_chunk, got end of stream') from e
    if response.WhichOneof('payload') != 'tensor_chunk':
      raise ValueError(
          f'Expected tensor_chunk, got "{response.WhichOneof("payload")}"'
          ' payload'
      )
    yield response.tensor_chunk


def _make_output_data(
    output: dna_model_pb2.Output,
    responses: Iterator[
        PredictResponse | dna_model_service_pb2.PredictVariantResponse
    ],
    interval: genome.Interval | None = None,
) -> track_data.TrackData | np.ndarray | junction_data.JunctionData:
  """Helper to create an output data object from an output proto."""
  match output.WhichOneof('payload'):
    case 'track_data':
      return track_data_utils.from_protos(
          output.track_data,
          _read_tensor_chunks(responses, output.track_data.values.chunk_count),
          interval=interval,
      )
    case 'data':
      chunks = _read_tensor_chunks(responses, output.data.chunk_count)
      values = tensor_utils.unpack_proto(output.data, chunks)
      return tensor_utils.upcast_floating(values)
    case 'junction_data':
      return junction_data_utils.from_protos(
          output.junction_data,
          _read_tensor_chunks(
              responses, output.junction_data.values.chunk_count
          ),
          interval=interval,
      )
    case _:
      raise ValueError(
          f'Unsupported output type: {output.WhichOneof("payload")}'
      )


def construct_output_metadata(
    responses: Iterator[dna_model_service_pb2.MetadataResponse],
) -> OutputMetadata:
  """Constructs an OutputMetadata from a stream of responses."""
  metadata = {}
  for response in responses:
    for metadata_proto in response.output_metadata:
      match metadata_proto.WhichOneof('payload'):
        case 'tracks':
          metadata[metadata_proto.output_type] = (
              track_data_utils.metadata_from_proto(metadata_proto.tracks)
          )
        case 'junctions':
          metadata[metadata_proto.output_type] = (
              junction_data_utils.metadata_from_proto(metadata_proto.junctions)
          )
        case _:
          raise ValueError(
              'Unsupported metadata type:'
              f' {metadata_proto.WhichOneof("payload")}'
          )

  return OutputMetadata(
      atac=metadata.get(dna_model_pb2.OUTPUT_TYPE_ATAC),
      cage=metadata.get(dna_model_pb2.OUTPUT_TYPE_CAGE),
      dnase=metadata.get(dna_model_pb2.OUTPUT_TYPE_DNASE),
      rna_seq=metadata.get(dna_model_pb2.OUTPUT_TYPE_RNA_SEQ),
      chip_histone=metadata.get(dna_model_pb2.OUTPUT_TYPE_CHIP_HISTONE),
      chip_tf=metadata.get(dna_model_pb2.OUTPUT_TYPE_CHIP_TF),
      splice_sites=metadata.get(dna_model_pb2.OUTPUT_TYPE_SPLICE_SITES),
      splice_site_usage=metadata.get(
          dna_model_pb2.OUTPUT_TYPE_SPLICE_SITE_USAGE
      ),
      splice_junctions=metadata.get(dna_model_pb2.OUTPUT_TYPE_SPLICE_JUNCTIONS),
      contact_maps=metadata.get(dna_model_pb2.OUTPUT_TYPE_CONTACT_MAPS),
      procap=metadata.get(dna_model_pb2.OUTPUT_TYPE_PROCAP),
  )


def _construct_output(
    output_dict: Mapping[
        dna_model_pb2.OutputType,
        track_data.TrackData | np.ndarray | junction_data.JunctionData,
    ],
):
  """Helper to construct an Output dataclass from a mapping of output types."""
  output = Output(
      atac=output_dict.get(dna_model_pb2.OUTPUT_TYPE_ATAC),
      cage=output_dict.get(dna_model_pb2.OUTPUT_TYPE_CAGE),
      dnase=output_dict.get(dna_model_pb2.OUTPUT_TYPE_DNASE),
      rna_seq=output_dict.get(dna_model_pb2.OUTPUT_TYPE_RNA_SEQ),
      chip_histone=output_dict.get(dna_model_pb2.OUTPUT_TYPE_CHIP_HISTONE),
      chip_tf=output_dict.get(dna_model_pb2.OUTPUT_TYPE_CHIP_TF),
      splice_sites=output_dict.get(dna_model_pb2.OUTPUT_TYPE_SPLICE_SITES),
      splice_site_usage=output_dict.get(
          dna_model_pb2.OUTPUT_TYPE_SPLICE_SITE_USAGE
      ),
      splice_junctions=output_dict.get(
          dna_model_pb2.OUTPUT_TYPE_SPLICE_JUNCTIONS
      ),
      contact_maps=output_dict.get(dna_model_pb2.OUTPUT_TYPE_CONTACT_MAPS),
      procap=output_dict.get(dna_model_pb2.OUTPUT_TYPE_PROCAP),
  )
  return output


def _make_output(
    responses: Iterator[PredictResponse],
    *,
    interval: genome.Interval | None = None,
) -> Output:
  """Helper to load an Output dataclass from a stream of response protos."""
  outputs = {}

  for response in responses:
    match response.WhichOneof('payload'):
      case 'output':
        outputs[response.output.output_type] = _make_output_data(
            response.output, responses, interval=interval
        )
      case 'tensor_chunk':
        raise ValueError('Received tensor chunk before output proto.')
      case _:
        raise ValueError(
            f'Unsupported response type: {response.WhichOneof("payload")}'
        )

  return _construct_output(outputs)


def _make_variant_output(
    responses: Iterator[dna_model_service_pb2.PredictVariantResponse],
) -> VariantOutput:
  """Helper to load an Output dataclass from a stream of response protos."""
  ref_outputs = {}
  alt_outputs = {}

  for response in responses:
    match response.WhichOneof('payload'):
      case 'reference_output':
        ref_outputs[response.reference_output.output_type] = _make_output_data(
            response.reference_output, responses
        )
      case 'alternate_output':
        alt_outputs[response.alternate_output.output_type] = _make_output_data(
            response.alternate_output, responses
        )
      case 'tensor_chunk':
        raise ValueError('Received tensor chunk before output proto.')
      case _:
        raise ValueError(
            f'Unsupported response type: {response.WhichOneof("payload")}'
        )

  return VariantOutput(
      reference=_construct_output(ref_outputs),
      alternate=_construct_output(alt_outputs),
  )


def _construct_anndata_from_proto(
    scorer_metadata: Sequence[dna_model_pb2.GeneScorerMetadata] | None,
    track_metadata: Sequence[dna_model_pb2.TrackMetadata],
    values: np.ndarray,
    interval: genome.Interval,
    variant: dna_model_pb2.Variant | None = None,
) -> anndata.AnnData:
  """Helper to construct `AnnData` from a scoring proto."""
  obs = None
  if scorer_metadata:
    metadata = []
    for gene_proto in scorer_metadata:
      scorer_metadata = {
          'gene_id': gene_proto.gene_id,
      }
      if gene_proto.HasField('strand'):
        scorer_metadata['strand'] = str(
            genome.Strand.from_proto(gene_proto.strand)
        )

      if gene_proto.HasField('name'):
        scorer_metadata['gene_name'] = gene_proto.name
      if gene_proto.HasField('type'):
        scorer_metadata['gene_type'] = gene_proto.type
      if gene_proto.HasField('junction_start'):
        scorer_metadata['junction_Start'] = gene_proto.junction_start
      if gene_proto.HasField('junction_end'):
        scorer_metadata['junction_End'] = gene_proto.junction_end

      metadata.append(scorer_metadata)

    obs = pd.DataFrame(metadata)
    obs.index = obs.index.map(str)

  var = track_data_utils.metadata_from_proto(
      dna_model_pb2.TracksMetadata(metadata=track_metadata)
  )
  var.index = var.index.map(str)

  uns = {'interval': interval}
  if variant is not None:
    uns['variant'] = genome.Variant.from_proto(variant)
  layers = None
  if values.shape[0] > 1:
    layers = {'quantiles': values[1]}

  return anndata.AnnData(
      X=values[0],
      obs=obs,
      var=var,
      uns=uns,
      layers=layers,
  )


def _construct_score_variant(
    output: dna_model_pb2.ScoreVariantOutput,
    responses: Iterator[
        dna_model_service_pb2.ScoreVariantResponse
        | dna_model_service_pb2.ScoreIsmVariantResponse
    ],
    interval: genome.Interval,
) -> anndata.AnnData:
  """Returns `AnnData`  of variant scores from a ScoreVariantOutput proto."""
  # dna_model_pb2.ScoreVariantOutput currently has ONLY VariantData with values
  # and metadata.
  chunks = _read_tensor_chunks(
      responses, output.variant_data.values.chunk_count
  )
  values = tensor_utils.unpack_proto(output.variant_data.values, chunks)
  values = tensor_utils.upcast_floating(values)
  anndata_scores = _construct_anndata_from_proto(
      output.variant_data.metadata.gene_metadata,
      output.variant_data.metadata.track_metadata,
      values,
      interval=interval,
      variant=output.variant_data.metadata.variant,
  )
  return anndata_scores


def _make_score_variant_output(
    responses: Iterator[
        dna_model_service_pb2.ScoreVariantResponse
        | dna_model_service_pb2.ScoreIsmVariantResponse
    ],
    interval: genome.Interval,
) -> list[anndata.AnnData]:
  """Returns a list of `AnnData` variant scores from a stream of responses."""

  anndata_scores = []
  for response in responses:
    match response.WhichOneof('payload'):
      case 'output':
        anndata_scores.append(
            _construct_score_variant(
                response.output, responses, interval=interval
            )
        )
      case 'tensor_chunk':
        raise ValueError('Received tensor chunk before output proto.')
      case _:
        raise ValueError(
            f'Unsupported response type: {response.WhichOneof("payload")}'
        )
  return anndata_scores


def _construct_score_interval(
    output: dna_model_pb2.ScoreIntervalOutput,
    responses: Iterator[dna_model_service_pb2.ScoreIntervalResponse],
    interval: genome.Interval,
) -> anndata.AnnData:
  """Returns `AnnData` of interval scores from a ScoreIntervalOutput proto."""
  chunks = _read_tensor_chunks(
      responses, output.interval_data.values.chunk_count
  )
  values = tensor_utils.unpack_proto(output.interval_data.values, chunks)
  values = tensor_utils.upcast_floating(values)
  anndata_scores = _construct_anndata_from_proto(
      output.interval_data.metadata.gene_metadata,
      output.interval_data.metadata.track_metadata,
      values,
      interval=interval,
  )
  return anndata_scores


def _make_interval_output(
    responses: Iterator[dna_model_service_pb2.ScoreIntervalResponse],
    interval: genome.Interval,
) -> list[anndata.AnnData]:
  """Returns a list of `AnnData` interval scores from a stream of responses."""

  anndata_scores = []
  for response in responses:
    match response.WhichOneof('payload'):
      case 'output':
        anndata_scores.append(
            _construct_score_interval(
                response.output,
                responses,
                interval=interval,
            )
        )
      case 'tensor_chunk':
        raise ValueError('Received tensor chunk before output proto.')
      case _:
        raise ValueError(
            f'Unsupported response type: {response.WhichOneof("payload")}'
        )
  return anndata_scores


def _convert_ontologies_to_protos(
    ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
) -> Sequence[dna_model_pb2.OntologyTerm] | None:
  """Convert ontology terms or curies to unique list of OntologyTerm protos."""
  if ontology_terms is None:
    return None
  else:
    protos = []
    for term in dict.fromkeys(ontology_terms):
      if isinstance(term, ontology.OntologyTerm):
        protos.append(term.to_proto())
      else:
        protos.append(ontology.from_curie(term).to_proto())
    return protos


def validate_sequence_length(length: int):
  """Validate that the input sequence length is supported by the model."""
  if length not in SUPPORTED_SEQUENCE_LENGTHS.values():
    raise ValueError(
        f'Sequence length {length} not supported by the model.'
        f' Supported lengths: {[*SUPPORTED_SEQUENCE_LENGTHS.values()]}'
    )


class DnaClient(dna_model.DnaModel):
  """Client for interacting with a DNA model server.

  Attributes:
    channel: gRPC channel to the DNA model server.
    metadata: Metadata to send with each request.
    model_version: Optional model version to use for the DNA model server. If
      none provided, the default model version will be used.
  """

  def __init__(
      self,
      *,
      channel: grpc.Channel,
      model_version: ModelVersion | None = None,
      metadata: Sequence[tuple[str, str]] = (),
  ):
    self._channel = channel
    self._metadata = metadata
    self._model_version = (
        model_version.name if model_version is not None else None
    )

  @retry_rpc
  def predict_sequence(
      self,
      sequence: str,
      *,
      organism: Organism = Organism.HOMO_SAPIENS,
      requested_outputs: Iterable[OutputType],
      ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
      interval: genome.Interval | None = None,
  ) -> Output:
    """Generate predictions for a given DNA sequence.

    Args:
      sequence: DNA sequence to make prediction for.
      organism: Organism to use for the prediction.
      requested_outputs: Iterable of OutputTypes indicating which subsets of
        predictions to return.
      ontology_terms: Iterable of ontology terms or curies to generate
        predictions for. If None returns all ontologies.
      interval: Optional interval from which the sequence was derived. This is
        used as the interval in the output TrackData.

    Returns:
      Output for the provided DNA sequence.
    """
    if not (unique_values := set(sequence)).issubset(
        _VALID_SEQUENCE_CHARACTERS
    ):
      invalid_characters = ','.join(unique_values - _VALID_SEQUENCE_CHARACTERS)
      raise ValueError(
          'Invalid sequence, must only contain the characters "ACGTN". Found'
          f' invalid characters: "{invalid_characters}"'
      )
    validate_sequence_length(len(sequence))
    requested_outputs = [o.to_proto() for o in dict.fromkeys(requested_outputs)]
    request = dna_model_service_pb2.PredictSequenceRequest(
        sequence=sequence,
        organism=organism.to_proto(),
        ontology_terms=_convert_ontologies_to_protos(ontology_terms),
        requested_outputs=requested_outputs,
        model_version=self._model_version,
    )
    responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
        self._channel
    ).PredictSequence(iter([request]), metadata=self._metadata)
    return _make_output(responses, interval=interval)

  @retry_rpc
  def predict_interval(
      self,
      interval: genome.Interval,
      *,
      organism: Organism = Organism.HOMO_SAPIENS,
      requested_outputs: Iterable[OutputType],
      ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
  ) -> Output:
    """Generate predictions for a given DNA interval.

    Args:
      interval: DNA interval to make prediction for.
      organism: Organism to use for the prediction.
      requested_outputs: Iterable of OutputTypes indicating which subsets of
        predictions to return.
      ontology_terms: Iterable of ontology terms or curies to generate
        predictions for. If None returns all ontologies.

    Returns:
      Output for the provided DNA interval.
    """
    validate_sequence_length(interval.width)
    requested_outputs = [o.to_proto() for o in dict.fromkeys(requested_outputs)]
    request = dna_model_service_pb2.PredictIntervalRequest(
        interval=interval.to_proto(),
        organism=organism.to_proto(),
        ontology_terms=_convert_ontologies_to_protos(ontology_terms),
        requested_outputs=requested_outputs,
        model_version=self._model_version,
    )
    responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
        self._channel
    ).PredictInterval(iter([request]), metadata=self._metadata)
    return _make_output(responses)

  @retry_rpc
  def predict_variant(
      self,
      interval: genome.Interval,
      variant: genome.Variant,
      *,
      organism: Organism = Organism.HOMO_SAPIENS,
      requested_outputs: Iterable[OutputType],
      ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
  ) -> VariantOutput:
    """Generate predictions for a given DNA variant.

    Args:
      interval: DNA interval to make prediction for.
      variant: DNA variant to make prediction for.
      organism: Organism to use for the prediction.
      requested_outputs: Iterable of OutputTypes indicating which subsets of
        predictions to return.
      ontology_terms: Iterable of ontology terms or curies to generate
        predictions for. If None returns all ontologies.

    Returns:
      Variant output for the provided DNA interval and variant.
    """
    validate_sequence_length(interval.width)
    requested_outputs = [o.to_proto() for o in dict.fromkeys(requested_outputs)]
    request = dna_model_service_pb2.PredictVariantRequest(
        interval=interval.to_proto(),
        variant=variant.to_proto(),
        organism=organism.to_proto(),
        ontology_terms=_convert_ontologies_to_protos(ontology_terms),
        requested_outputs=requested_outputs,
        model_version=self._model_version,
    )
    responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
        self._channel
    ).PredictVariant(iter([request]), metadata=self._metadata)
    return _make_variant_output(responses)

  @retry_rpc
  def score_interval(
      self,
      interval: genome.Interval,
      interval_scorers: Sequence[interval_scorers_lib.IntervalScorerTypes] = (),
      *,
      organism: Organism = Organism.HOMO_SAPIENS,
  ) -> list[anndata.AnnData]:
    """Generate interval scores for a single given interval.

    Args:
      interval: Interval to make prediction for.
      interval_scorers: Sequence of interval scorers to use for scoring. If no
        interval scorers are provided, the recommended interval scorers for the
        organism will be used.
      organism: Organism to use for the prediction.

    Returns:
      List of `AnnData` interval scores.
    """
    if not interval_scorers:
      interval_scorers = list(
          interval_scorers_lib.RECOMMENDED_INTERVAL_SCORERS.values()
      )

    validate_sequence_length(interval.width)
    if len(interval_scorers) != len(set(interval_scorers)):
      raise ValueError(
          f'Duplicate interval scorers requested: {interval_scorers}.'
      )
    if len(interval_scorers) > MAX_VARIANT_SCORERS_PER_REQUEST:
      raise ValueError(
          'Too many interval scorers requested, '
          f'maximum allowed: {MAX_VARIANT_SCORERS_PER_REQUEST}.'
      )
    if any(
        scorer.width is not None and scorer.width > interval.width
        for scorer in interval_scorers
    ):
      raise ValueError('Interval scorers must have widths <= interval width.')
    request = dna_model_service_pb2.ScoreIntervalRequest(
        interval=interval.to_proto(),
        interval_scorers=[scorer.to_proto() for scorer in interval_scorers],
        organism=organism.to_proto(),
        model_version=self._model_version,
    )
    responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
        self._channel
    ).ScoreInterval(iter([request]), metadata=self._metadata)
    interval_outputs = _make_interval_output(responses, interval)
    for ix in range(len(interval_scorers)):
      interval_outputs[ix].uns['interval_scorer'] = interval_scorers[ix]
    return interval_outputs

  @retry_rpc
  def score_variant(
      self,
      interval: genome.Interval,
      variant: genome.Variant,
      variant_scorers: Sequence[variant_scorers_lib.VariantScorerTypes] = (),
      *,
      organism: Organism = Organism.HOMO_SAPIENS,
  ) -> list[anndata.AnnData]:
    """Generate variant scores for a single given DNA variant.

    Args:
      interval: DNA interval to make prediction for.
      variant: DNA variant to make prediction for.
      variant_scorers: Sequence of variant scorers to use for scoring. If no
        variant scorers are provided, the recommended variant scorers for the
        organism will be used.
      organism: Organism to use for the prediction.

    Returns:
      List of `AnnData` variant scores.
    """
    if not variant_scorers:
      variant_scorers = variant_scorers_lib.get_recommended_scorers(
          organism.to_proto()
      )

    validate_sequence_length(interval.width)
    for variant_scorer in variant_scorers:
      supported_organisms = variant_scorers_lib.SUPPORTED_ORGANISMS[
          variant_scorer.base_variant_scorer
      ]
      if organism.to_proto() not in supported_organisms:
        supported_organisms = [
            dna_model_pb2.Organism.Name(o).removeprefix('ORGANISM_')
            for o in supported_organisms
        ]
        raise ValueError(
            f'Unsupported organism: {organism.name} for scorer'
            f' {variant_scorer}. Supported organisms:'
            f' {supported_organisms}'
        )
    if len(variant_scorers) != len(set(variant_scorers)):
      raise ValueError(
          f'Duplicate variant scorers requested: {variant_scorers}.'
      )
    if len(variant_scorers) > MAX_VARIANT_SCORERS_PER_REQUEST:
      raise ValueError(
          'Too many variant scorers requested, '
          f'maximum allowed: {MAX_VARIANT_SCORERS_PER_REQUEST}.'
      )
    request = dna_model_service_pb2.ScoreVariantRequest(
        interval=interval.to_proto(),
        variant=variant.to_proto(),
        organism=organism.to_proto(),
        variant_scorers=[vs.to_proto() for vs in variant_scorers],
        model_version=self._model_version,
    )
    responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
        self._channel
    ).ScoreVariant(iter([request]), metadata=self._metadata)
    variant_outputs = _make_score_variant_output(responses, interval)
    for ix in range(len(variant_scorers)):
      variant_outputs[ix].uns['variant_scorer'] = variant_scorers[ix]
    return variant_outputs

  def score_ism_variants(
      self,
      interval: genome.Interval,
      ism_interval: genome.Interval,
      variant_scorers: Sequence[variant_scorers_lib.VariantScorerTypes] = (),
      *,
      organism: Organism = Organism.HOMO_SAPIENS,
      progress_bar: bool = True,
      max_workers: int = dna_model.DEFAULT_MAX_WORKERS,
  ) -> list[list[anndata.AnnData]]:
    """Generate in-silico mutagenesis (ISM) variant scores for a given interval.

    Args:
      interval: DNA interval to make the prediction for.
      ism_interval: Interval to perform ISM.
      variant_scorers: Sequence of variant scorers to use for scoring each
        variant. If no variant scorers are provided, the recommended variant
        scorers for the organism will be used.
      organism: Organism to use for the prediction.
      progress_bar: If True, show a progress bar.
      max_workers: Number of parallel workers to use.

    Returns:
      List of variant scores for each variant in the ISM interval.
    """
    if not variant_scorers:
      variant_scorers = variant_scorers_lib.get_recommended_scorers(
          organism.to_proto()
      )

    validate_sequence_length(interval.width)
    if ism_interval.negative_strand:
      raise ValueError('ISM interval must be on the positive strand.')

    ism_intervals = []
    for position in range(0, ism_interval.width, MAX_ISM_INTERVAL_WIDTH):
      ism_intervals.append(
          genome.Interval(
              ism_interval.chromosome,
              ism_interval.start + position,
              min(
                  ism_interval.start + position + MAX_ISM_INTERVAL_WIDTH,
                  ism_interval.end,
              ),
          )
      )

    @retry_rpc
    def _score_ism_variant(
        ism_interval: genome.Interval,
    ) -> list[anndata.AnnData]:
      request = dna_model_service_pb2.ScoreIsmVariantRequest(
          interval=interval.to_proto(),
          ism_interval=ism_interval.to_proto(),
          organism=organism.to_proto(),
          variant_scorers=[vs.to_proto() for vs in variant_scorers],
          model_version=self._model_version,
      )
      responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
          self._channel
      ).ScoreIsmVariant(iter([request]), metadata=self._metadata)

      return _make_score_variant_output(responses, interval)

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=max_workers
    ) as executor:
      futures = [
          executor.submit(_score_ism_variant, ism_interval)
          for ism_interval in ism_intervals
      ]
      ism_scores = []
      for future in tqdm.auto.tqdm(
          concurrent.futures.as_completed(futures),
          total=len(futures),
          disable=not progress_bar,
      ):
        ism_scores.extend(future.result())

    # Group scores by variant.
    ism_scores_by_variant: dict[str, list[anndata.AnnData]] = {}
    for variant_score in ism_scores:
      ism_scores_by_variant.setdefault(
          str(variant_score.uns['variant']), []
      ).append(variant_score)

    return list(ism_scores_by_variant.values())

  def output_metadata(
      self, organism: Organism = Organism.HOMO_SAPIENS
  ) -> OutputMetadata:
    """Get the metadata for a given organism.

    Args:
      organism: Organism to get metadata for.

    Returns:
      OutputMetadata for the provided organism.
    """
    request = dna_model_service_pb2.MetadataRequest(
        organism=organism.to_proto()
    )
    responses = dna_model_service_pb2_grpc.DnaModelServiceStub(
        self._channel
    ).GetMetadata(request, metadata=self._metadata)
    return construct_output_metadata(responses)


def create(
    api_key: str,
    *,
    model_version: ModelVersion | None = None,
    timeout: float | None = None,
    address: str | None = None,
) -> DnaClient:
  """Creates a model client for a given API key.

  Args:
    api_key: API key to use for authentication.
    model_version: Optional model version to use for the DNA model server. If
      none is provided, the default model will be used.
    timeout: Optional timeout for waiting for the channel to be ready.
    address: Optional server address to connect to.

  Returns:
   `DnaClient` instance.
  """
  options = [
      ('grpc.max_send_message_length', -1),
      ('grpc.max_receive_message_length', -1),
  ]
  address = address or 'dns:///gdmscience.googleapis.com:443'
  channel = grpc.secure_channel(
      address, grpc.ssl_channel_credentials(), options=options
  )
  grpc.channel_ready_future(channel).result(timeout)

  return DnaClient(
      channel=channel,
      model_version=model_version,
      metadata=[('x-goog-api-key', api_key)],
  )



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for converting JunctionData to/from protos."""

from collections.abc import Iterable, Sequence

from alphagenome import tensor_utils
from alphagenome.data import genome
from alphagenome.data import junction_data
from alphagenome.data import ontology
from alphagenome.protos import dna_model_pb2
from alphagenome.protos import tensor_pb2
import numpy as np
import pandas as pd


def to_protos(
    data: junction_data.JunctionData,
    *,
    bytes_per_chunk: int = 0,
    compression_type: tensor_pb2.CompressionType = (
        tensor_pb2.CompressionType.COMPRESSION_TYPE_NONE
    ),
) -> tuple[dna_model_pb2.JunctionData, Sequence[tensor_pb2.TensorChunk]]:
  """Converts the `JunctionData` to protobuf messages.

  Args:
    data: The `JunctionData` object to convert to protos.
    bytes_per_chunk: The maximum number of bytes per tensor chunk.
    compression_type: The compression type to use for the tensor chunks.

  Returns:
    A tuple containing the `JunctionData` protobuf message and a sequence of
      `TensorChunk` protobuf messages.
  """
  tensor, chunks = tensor_utils.pack_tensor(
      data.values,
      bytes_per_chunk=bytes_per_chunk,
      compression_type=compression_type,
  )

  return (
      dna_model_pb2.JunctionData(
          junctions=[j.to_proto() for j in data.junctions],
          values=tensor,
          metadata=metadata_to_proto(data.metadata).metadata,
          interval=data.interval.to_proto() if data.interval else None,
      ),
      chunks,
  )


def from_protos(
    proto: dna_model_pb2.JunctionData,
    chunks: Iterable[tensor_pb2.TensorChunk] = (),
    *,
    interval: genome.Interval | None = None,
) -> junction_data.JunctionData:
  """Converts a `JunctionData` protobuf to a `JunctionData` object.

  Args:
    proto: A `JunctionData` protobuf message.
    chunks: A sequence of `TensorChunk` protobuf messages.
    interval: Optional `Interval` object representing the genomic region
      containing the junctions. Only used if the proto does not have an
      interval.

  Returns:
    A `JunctionData` object.
  """
  values = tensor_utils.unpack_proto(proto.values, chunks)
  values = tensor_utils.upcast_floating(values)

  metadata = metadata_from_proto(
      dna_model_pb2.JunctionsMetadata(metadata=proto.metadata)
  )

  if proto.HasField('interval'):
    interval = genome.Interval.from_proto(proto.interval)

  return junction_data.JunctionData(
      junctions=np.array(
          [genome.Interval.from_proto(j) for j in proto.junctions]
      ),
      values=values,
      metadata=metadata,
      interval=interval,
  )


def metadata_to_proto(
    metadata: junction_data.JunctionMetadata,
) -> dna_model_pb2.JunctionsMetadata:
  """Converts junction metadata to a JunctionsMetadata.

  Args:
    metadata: A pandas DataFrame containing junction metadata.

  Returns:
    A `JunctionsMetadata` protobuf message.
  """
  names = metadata['name']
  default_values = [None] * len(names)

  columns = zip(
      metadata['name'],
      metadata.get('ontology_curie', default_values),
      metadata.get('biosample_type', default_values),
      metadata.get('biosample_name', default_values),
      metadata.get('biosample_life_stage', default_values),
      metadata.get('gtex_tissue', default_values),
      metadata.get('data_source', default_values),
      metadata.get('Assay title', default_values),
      strict=True,
  )

  metadata_protos = []

  for (
      name,
      ontology_curie,
      biosample_type,
      biosample_name,
      biosample_life_stage,
      gtex_tissue,
      data_source,
      assay,
  ) in columns:
    if biosample_type is not None:
      biosample_proto = dna_model_pb2.Biosample(
          name=biosample_name,
          type=dna_model_pb2.BiosampleType.Value(
              f'BIOSAMPLE_TYPE_{biosample_type.upper()}'
          ),
          stage=biosample_life_stage,
      )
    else:
      biosample_proto = None

    metadata_protos.append(
        dna_model_pb2.JunctionMetadata(
            name=name,
            ontology_term=ontology.from_curie(ontology_curie).to_proto()
            if ontology_curie
            else None,
            biosample=biosample_proto,
            gtex_tissue=gtex_tissue,
            data_source=data_source,
            assay=assay,
        )
    )

  return dna_model_pb2.JunctionsMetadata(metadata=metadata_protos)


def metadata_from_proto(
    proto: dna_model_pb2.JunctionsMetadata,
) -> junction_data.JunctionMetadata:
  """Create JunctionMetadata from a dna_model_pb2.JunctionsMetadata.

  Args:
    proto: A `JunctionsMetadata` protobuf message.

  Returns:
    A pandas DataFrame containing junction metadata.
  """
  metadata = []
  for junction_proto in proto.metadata:
    junction_metadata = {
        'name': junction_proto.name,
    }

    if junction_proto.HasField('ontology_term'):
      junction_metadata['ontology_curie'] = ontology.from_proto(
          junction_proto.ontology_term
      ).ontology_curie

    if junction_proto.HasField('biosample'):
      junction_metadata['biosample_name'] = junction_proto.biosample.name
      junction_metadata['biosample_type'] = (
          dna_model_pb2.BiosampleType.Name(junction_proto.biosample.type)
          .removeprefix('BIOSAMPLE_TYPE_')
          .lower()
      )
      if junction_proto.biosample.HasField('stage'):
        junction_metadata['biosample_life_stage'] = (
            junction_proto.biosample.stage
        )

    if junction_proto.HasField('gtex_tissue'):
      junction_metadata['gtex_tissue'] = junction_proto.gtex_tissue

    if junction_proto.HasField('data_source'):
      junction_metadata['data_source'] = junction_proto.data_source

    if junction_proto.HasField('assay'):
      junction_metadata['Assay title'] = junction_proto.assay

    metadata.append(junction_metadata)

  if metadata:
    return pd.DataFrame(metadata)
  else:
    return pd.DataFrame(columns=['name'])



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualize transcripts/gene annotation in matplotlib."""

from collections.abc import Sequence
import dataclasses
import enum
import math
from typing import Any

from alphagenome.data import genome
from alphagenome.data import transcript as transcript_utils
import intervaltree
import matplotlib as mpl
import matplotlib.pyplot as plt


@dataclasses.dataclass
class TranscriptStyle:
  """Style specification for a transcript plot.

  CDS = protein coding sequence.
  UTR = untranslated region.

  Attributes:
    cds_height: CDS height.
    utr_height: UTR height.
    cds_color: CDS color.
    utr5_color: 5' UTR color.
    utr3_color: 3' UTR color.
    first_noncoding_exon_color: Color of the first non-coding exon. This helps
      to indicate transcript directionality.
    label_color: Label color.
    xlim_pad: Controls amount of whitespace on the region flanks.
  """

  cds_height: float
  utr_height: float
  cds_color: str
  utr5_color: str
  utr3_color: str
  first_noncoding_exon_color: str
  label_color: str
  xlim_pad: float


class TranscriptStylePreset(enum.Enum):
  """Style enum for transcript plots.

  Attributes:
    STANDARD: Standard transcript style.
    MINIMAL: Minimal transcript style.
  """

  STANDARD = TranscriptStyle(
      cds_height=0.7,
      utr_height=0.35,
      cds_color='#7f7f7f',  # Grey from tab10 palette.
      utr5_color='#ff7f0e',  # Orange.
      utr3_color='#1f77b4',  # Blue.
      first_noncoding_exon_color='#2ca02c',  # Green.
      label_color='#7f7f7f',
      xlim_pad=0.01,
  )

  MINIMAL = TranscriptStyle(
      cds_height=0.4,
      utr_height=0.22,
      cds_color='black',
      utr5_color='black',
      utr3_color='black',
      # TODO(b/377291432): find a nice way of specifying strand orientation.
      first_noncoding_exon_color='black',
      label_color='black',
      xlim_pad=0.05,
  )


def plot_transcripts(
    ax: plt.Axes,
    transcripts: Sequence[transcript_utils.Transcript],
    interval: genome.Interval,
    zero_origin: bool = False,
    label_name: str | None = None,
    transcript_style: TranscriptStyle = TranscriptStylePreset.STANDARD.value,
    plot_labels_once: bool = False,
    **kwargs,
) -> mpl.figure.Figure:
  """Plot transcripts.

  Loops over each transcript in `transcripts` and calls `draw_transcript`.

  Args:
    ax: Matplotlib axis onto which to plot transcript annotations.
    transcripts: Sequence of transcripts returned by a
      transcript.TranscriptExtractor.
    interval: Genomic interval at which to visualize the transcripts.
    zero_origin: If True, the beginning of the interval will start with 0.
    label_name: Which label in transcript.info to draw next to the transcript.
    transcript_style: specification of transcript styling details.
    plot_labels_once: If True, labels will only be plotted once.
    **kwargs: kwargs passed to draw_transcript.

  Returns:
    Matplotlib figure object.
  """
  if not transcripts:
    return

  # Slightly pad x limits for nicer spacing, and shift x axis limits if needed.
  shift = -interval.start if zero_origin else 0
  xlim_pad = interval.width * transcript_style.xlim_pad
  ax.set_xlim(
      [interval.start + shift - xlim_pad, interval.end + shift + xlim_pad]
  )

  # Get typical label width and transcript heights.
  text_width = _get_text_width(transcripts[0].info[label_name], ax=ax)
  heights = _get_placement_heights(
      transcripts, extend_fraction=1.0, front_padding=text_width
  )

  labels_already_drawn = []
  for transcript in transcripts:
    # Add transcript labels.
    if label_name is not None:
      label = transcript.info[label_name]
    else:
      label = None

    draw_transcript(
        ax=ax,
        transcript=transcript,
        interval=interval,
        y=heights[transcript.transcript_id],
        cds_height=transcript_style.cds_height,
        utr_height=transcript_style.utr_height,
        cds_color=transcript_style.cds_color,
        utr5_color=transcript_style.utr5_color,
        utr3_color=transcript_style.utr3_color,
        first_noncoding_exon_color=transcript_style.first_noncoding_exon_color,
        label_color=transcript_style.label_color,
        shift=shift,
        label=None
        if (label in labels_already_drawn and plot_labels_once)
        else label,
        **kwargs,
    )

    labels_already_drawn.append(label)

  ax.set_ylim([min(heights.values()) - 1, max(heights.values()) + 1])


def draw_transcript(
    ax: plt.Axes,
    transcript: transcript_utils.Transcript,
    interval: genome.Interval,
    y: float,
    cds_height: float = 0.7,
    utr_height: float = 0.35,
    cds_color: str = '#7f7f7f',  # Grey from tab10 palette.
    utr5_color: str = '#ff7f0e',  # Orange.
    utr3_color: str = '#1f77b4',  # Blue.
    first_noncoding_exon_color: str = '#2ca02c',  # Green.
    shift: int = 0,
    label: str | None = None,
    label_color: str = '#7f7f7f',
    **kwargs,
) -> None:
  """Draw an individual transcript as rectangular components on an axis.

  CDS = protein coding sequence.
  UTR = untranslated region.

  This function is used by `plot_transcripts`.

  Args:
    ax: Matplotlib axis onto which to draw the transcript.
    transcript: Transcript to draw.
    interval: Genomic interval at which to visualize the transcript.
    y: Vertical position at which to draw the transcript.
    cds_height: CDS height.
    utr_height: UTR height.
    cds_color: CDS color in hex string format.
    utr5_color: 5' UTR color in hex string format.
    utr3_color: 3' UTR color  in hex string format.
    first_noncoding_exon_color: Color of the first non-coding exon. This helps
      to indicate transcript directionality. Hex string format.
    shift: X-axis shift.
    label: Optional label to draw next to the transcript.
    label_color: Label color.
    **kwargs: Additional keyword arguments passed to matplotlib plotting
      functions.
  """
  ax.set_yticklabels([])
  ax.set_yticks([])

  def draw_exons_and_introns(exons, color, exon_height):
    if not exons:
      return
    # 1. Draw all exons.
    for exon in exons:
      # TODO: b/377291432 - Skip drawing an exon if it will be drawn below
      # separately to avoid overlap if alpha<1 and overlaps in vector format.
      draw_interval(
          ax=ax,
          interval=exon,
          y=y,
          shift=shift,
          height=exon_height,
          color=color,
          **kwargs,
      )

    # 2. Draw all introns.
    for intron in transcript_utils.Transcript(exons).introns:
      ax.plot([intron.start, intron.end], [y, y], color=color, linewidth=0.5)
      # Draw max_num_arrows arrows, at least one arrow per intron.
      intron_to_interval_fraction = intron.width / interval.width
      # If this fraction is too small we do not draw arrow
      if intron_to_interval_fraction < 0.01:
        continue
      max_num_arrows = 10
      num_arrows = min(
          math.ceil(intron_to_interval_fraction * max_num_arrows),
          max_num_arrows,
      )
      space = intron.width // num_arrows

      for i in range(num_arrows):
        arrow_start = intron.start - space // 2 + i * space
        # Arrows are at most 10bp from the exons.
        if (
            arrow_start + space - 10 > interval.end
            or arrow_start < interval.start + 10
        ):
          continue
        style = '<-' if transcript.is_negative_strand else '->'
        ax.annotate(
            '',
            xy=(arrow_start + space, y),
            xytext=(arrow_start + space - 0.001, y),
            arrowprops=dict(
                arrowstyle=f'{style},head_width=0.3', linewidth=0.5, color=color
            ),
        )

  # First draw all exons and introns with UTR height.
  draw_exons_and_introns(
      transcript.exons, color=cds_color, exon_height=utr_height
  )
  draw_interval(
      ax=ax,
      interval=transcript.exons[0],
      y=y,
      shift=shift,
      label=label,
      height=utr_height,
      color=cds_color,
      label_color=label_color,
      **kwargs,
  )

  # Draw the first non-coding exon with a special color.
  first_exon_index = 0 if transcript.is_negative_strand else -1
  draw_interval(
      ax=ax,
      interval=transcript.exons[first_exon_index],
      y=y,
      height=utr_height,
      shift=shift,
      color=first_noncoding_exon_color,
      **kwargs,
  )

  # Add UTRs for coding transcripts.
  if transcript.cds is not None:
    draw_exons_and_introns(
        transcript.utr5, color=utr5_color, exon_height=utr_height
    )
    draw_exons_and_introns(
        transcript.cds, color=cds_color, exon_height=cds_height
    )
    draw_exons_and_introns(
        transcript.utr3, color=utr3_color, exon_height=utr_height
    )


def draw_interval(
    ax: plt.Axes,
    interval: genome.Interval,
    y: float,
    label: str | None = None,
    height: float = 0.5,
    shift: int = 0,
    label_color: str = '#7f7f7f',
    **kwargs,
):
  """Draw rectangle patch on the axis given a genomic interval.

  Args:
    ax: Matplotlib axis onto which to draw the interval.
    interval: Genomic interval to draw.
    y: Vertical position at which to draw the interval.
    label: Optional label to draw next to the interval.
    height: Height of the interval.
    shift: X-axis shift.
    label_color: Label color in hex string format.
    **kwargs: Additional keyword arguments passed to matplotlib plotting
      functions.
  """
  xy = (interval.start + shift, y - height / 2)
  ax.add_patch(
      mpl.patches.Rectangle(
          xy=xy,
          width=interval.width,
          height=height,
          clip_on=True,
          linewidth=0,
          **kwargs,
      )
  )

  # Add center-aligned text label.
  if label is not None:
    ax.text(
        x=max(xy[0], ax.get_xlim()[0]),
        y=y,
        s=label,
        color=label_color,
        horizontalalignment='right',
        verticalalignment='center',
    )


def _get_placement_heights(
    transcripts: Sequence[transcript_utils.Transcript],
    extend_fraction: float = 1.0,
    front_padding: float = 0.0,
) -> dict[Any, int]:
  """Get heights at which to place the transcripts."""
  # TODO: b/376672690 - Implement simpler packing algorithm.
  levels = [intervaltree.IntervalTree()]
  # Sort transcripts by length and start placing longest transcripts first.
  sorted_transcripts = sorted(
      transcripts, key=lambda x: x.transcript_interval.width, reverse=True
  )
  transcript_levels = {}
  for transcript in sorted_transcripts:
    placed = False
    level_idx = 0
    while not placed:
      if level_idx >= len(levels):
        levels.append(intervaltree.IntervalTree())
      if levels[level_idx].overlaps(
          transcript.transcript_interval.start - front_padding,
          transcript.transcript_interval.end,
      ):
        # Overlaps an existing interval -> increase the level.
        level_idx += 1
      else:
        # Doesn't overlap. Remember the interval.
        levels[level_idx].addi(
            transcript.transcript_interval.start - front_padding,
            int(transcript.transcript_interval.end * extend_fraction),
        )
        transcript_levels[transcript.transcript_id] = level_idx
        placed = True
  return {
      transcript_id: len(levels) - 1 - level_idx
      for transcript_id, level_idx in transcript_levels.items()
  }


def _get_text_width(label: str, ax: plt.Axes, **kwargs) -> float:
  """Get text width in data coordinates."""
  text = ax.text(0, 0, label, **kwargs)
  plt.gcf().canvas.draw()
  bb = text.get_window_extent().transformed(ax.transData.inverted())
  text.remove()  # Remove text.
  return bb.x1 - bb.x0



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from unittest import mock

from absl.testing import absltest
from alphagenome import colab_utils


_TEST_SECRET_KEY = '_TEST_ALPHAGENOME_API_KEY'


class ColabUtilsTest(absltest.TestCase):

  def test_get_api_key_from_environment(self):
    with mock.patch.dict(os.environ, {_TEST_SECRET_KEY: 'foo'}):
      self.assertEqual(colab_utils.get_api_key(_TEST_SECRET_KEY), 'foo')

  def test_get_api_key_from_environment_not_found_raises_error(self):
    with self.assertRaisesRegex(
        ValueError,
        f"Cannot find API key with secret='{_TEST_SECRET_KEY}'.",
    ):
      _ = colab_utils.get_api_key(_TEST_SECRET_KEY)

  def test_get_api_key_from_colab_secrets(self):
    mock_colab = mock.MagicMock()
    mock_colab.userdata.get.return_value = 'bar'

    with mock.patch.dict(
        sys.modules, {'google': mock.MagicMock(), 'google.colab': mock_colab}
    ):
      self.assertEqual(colab_utils.get_api_key(), 'bar')

  def test_get_api_key_from_colab_secrets_not_found_raises_error(self):
    mock_colab = mock.MagicMock()
    mock_colab.userdata.NotebookAccessError = Exception
    mock_colab.userdata.SecretNotFoundError = Exception
    mock_colab.userdata.TimeoutException = Exception

    mock_colab.userdata.get.side_effect = Exception()

    with mock.patch.dict(
        sys.modules, {'google': mock.MagicMock(), 'google.colab': mock_colab}
    ):
      secret = 'my_secret'
      with self.assertRaisesRegex(
          ValueError,
          f'Cannot find or access API key in Colab secrets with {secret=}.',
      ):
        _ = colab_utils.get_api_key(secret)


if __name__ == '__main__':
  absltest.main()
